searchNodes=[{"ref":"Mix.Tasks.Propcheck.html","title":"Mix.Tasks.Propcheck","type":"task","doc":"PropCheck runs property checking as part of ExUnit test and stores counter examples of failing properties in order to reapply them in the next test run. The file name for the counter examples can be configured in mix.exs in the project configuration as propcheck: [counter_example: &quot;filename&quot;] With mix propcheck.inspect you can inspect the found counter examples, with mix propcheck.clean the file is deleted afterwards."},{"ref":"Mix.Tasks.Propcheck.html#run/1","title":"Mix.Tasks.Propcheck.run/1","type":"function","doc":"A task needs to implement run which receives a list of command line args. Callback implementation for Mix.Task.run/1."},{"ref":"Mix.Tasks.Propcheck.Clean.html","title":"Mix.Tasks.Propcheck.Clean","type":"task","doc":"Removes the counter example file of propcheck."},{"ref":"Mix.Tasks.Propcheck.Inspect.html","title":"Mix.Tasks.Propcheck.Inspect","type":"task","doc":"Inspects all counter examples."},{"ref":"PropCheck.html","title":"PropCheck","type":"module","doc":"Provides the macros and functions for property based testing using proper as base implementation. PropCheck supports many features of PropEr, but the automated generation of test data generators is only partially supported due to internal features of PropEr focusing of Erlang only. Using PropCheck To use PropCheck, you need to add use PropCheck to your Elixir files. This gives you access to the functions and macros defined here as well as to the property macro, defined in PropCheck.Properties.property/4. To set default options for your properties, you can use use PropCheck, default_opts: [numtests: 50]. default_opts can be a list of options defined below or a function that returns a list of option. In most examples shown here, we directly use the quickcheck function, but typically you use the property macro instead to define test cases for ExUnit. Also available are the value generators which are imported directly from PropCheck.BasicTypes. How to write properties The simplest properties that PropEr can test consist of a single boolean expression (or a statement block that returns a boolean), which is expected to evaluate to true. Thus, the test true always succeeds, while the test false always fails (the failure of a property may also be signified by throwing an exception, error or exit. More complex (and useful) properties can be written by wrapping such a boolean expression with one or more of the following wrappers: forall/2 implies/2 when_fail/2 trap_exit/1 conjunction/1 equals/2 There are also multiple wrappers that can be used to collect statistics on the distribution of test data: collect/2 collect/3 aggregate/2 aggregate/3 classify/3 measure/3 A property may also be wrapped with one or more of the following outer-level wrappers, which control the behaviour of the testing subsystem. If an outer-level wrapper appears more than once in a property, the innermost instance takes precedence. numtests/2 fails/1 on_output/2 PropCheck follows the Elixir idioms that for fluent API the first parameter flows through a pipeline of functions. Therefore, in PropCheck the wrapper functions have the property as first argument allowing to use the |&gt; to concatenate wrapper functions. It helps to distinguish between the property to test and those wrappers which beautify the results or the collection information about the test values. This is a significant derivation of the API of both, PropEr and QuickCheck. For some actual usage examples, see the code in the examples directory, or check out PropEr&#39;s site. The testing modules in the tests directory may also be of interest. Program behaviour When running in verbose mode (this is the default for quickcheck), each successful test prints a . on screen. If a test fails, a ! is printed, along with the failing test case (the instances of the types in every forall) and the cause of the failure, if it was not simply the falsification of the property. Then, unless the test was expected to fail, PropEr attempts to produce a minimal test case that fails the property in the same way. This process is called shrinking. During shrinking, a . is printed for each successful simplification of the failing test case. When PropEr reaches its shrinking limit or realizes that the instance cannot be shrunk further while still failing the test, it prints the minimal failing test case and failure reason and exits. The return value of PropEr can be one of the following: true: The property held for all valid produced inputs. false: The property failed for some input. {error, type_of_error}: An error occurred; see the section Errors section for more information. To test all properties exported from a module (a property is a 0-arity function whose name begins with prop_), you can use module/1 or module/2. This returns a list of all failing properties, represented by MFAs. Testing progress is also printed on screen (unless quiet mode is active). The provided options are passed on to each property, except for long_result, which controls the return value format of the module function itself. Counterexamples A counterexample for a property is represented as a list of terms; each such term corresponds to the type in a forall. The instances are provided in the same order as the forall wrappers in the property, i.e. the instance at the head of the list corresponds to the outermost forall etc. Instances generated inside a failing sub-property of a conjunction are marked with the sub-property&#39;s tag. The last (simplest) counterexample produced by PropEr during a (failing) run can be retrieved after testing has finished, by running counterexample/0. When testing a whole module, run counterexamples/0 to get a counterexample for each failing property, as a list of {mfa, counterexample} tuples. To enable this functionality, some information has to remain in the process dictionary even after PropEr has returned. If, for some reason, you want to completely clean up the process dictionary of PropEr-produced entries, run clean_garbage/0. Counterexamples can also be retrieved by running PropEr in long-result mode, where counterexamples are returned as part of the return value. Specifically, when testing a single property under long-result mode (activated by supplying the option :long_result, or by calling counterexample/1 or counterexample/2 instead of quickcheck/1 and quickcheck/2 respectively), PropEr will return a counterexample in case of failure (instead of simply returning false). When testing a whole module under long-result mode (activated by supplying the option :long_result to module/2), PropEr will return a list of {mfa(), counterexample} tuples, one for each failing property. You can re-check a specific counterexample against the property that it previously falsified by running check/2 or check/3. This will return one of the following (both in short- and long-result mode): true: The property now holds for this test case. false: The test case still fails (although not necessarily for the same reason as before). {error, type_of_error}: An error occurred - see the section Errors section for more information. PropEr will not attempt to shrink the input in case it still fails the property. Unless silent mode is active, PropEr will also print a message on screen, describing the result of the re-checking. Note that PropEr can do very little to verify that the counterexample actually corresponds to the property that it is tested against. Options Options can be provided as an extra argument to most testing functions (such as quickcheck/1). A single option can be written stand-alone, or multiple options can be provided in a list. When two settings conflict, the one that comes first in the list takes precedence. Settings given inside external wrappers to a property (see the section on How to write properties) override any conflicting settings provided as options. The available options are: :quiet Enables quiet mode - no output is printed on screen while PropEr is running. :verbose Enables verbose mode - this is the default mode of operation. {:to_file, io_device} Redirects all of PropEr&#39;s output to io_device, which should be an IO device associated with a file opened for writing. {:on_output, output_function} PropEr will use the supplied function for all output printing. This function should accept two arguments in the style of :io.format/2. CAUTION: The above output control options are incompatible with each other. :long_result Enables long-result mode (see the section Counterexamples for details). {:numtests, positive_number} or simply positive_number This is equivalent to the numtests/1 property wrapper. Any numtests/1 wrappers in the actual property will overwrite this setting. {:start_size, size} Specifies the initial value of the size parameter (default is 1), see the documentation of the PropCheck.BasicTypes module for details. {:max_size, size} Specifies the maximum value of the size parameter (default is 42), see the documentation of the PropCheck.BasicTypes module for details. {:max_shrinks, non_negative_number} Specifies the maximum number of times a failing test case should be shrunk before returning. Note that the shrinking may stop before so many shrinks are achieved if the shrinking subsystem deduces that it cannot shrink the failing test case further. Default is 500. :noshrink Instructs PropEr to not attempt to shrink any failing test cases. {:constraint_tries, positive_number} Specifies the maximum number of tries before the generator subsystem gives up on producing an instance that satisfies a such_that constraint. Default is 50. fails This is equivalent to the fails/1 property wrapper. {:spec_timeout, :infinity | &lt;Non_negative_number&gt;} When testing a spec, PropEr will consider an input to be failing if the function under test takes more than the specified amount of milliseconds to return for that input. :any_to_integer All generated instances of the type PropCheck.BasicTypes.any/0 will be integers. This is provided as a means to speed up the testing of specs, where any is a commonly used type. Remark: PropCheck does not support spec-testing. {:search_steps, non_negative_number} used for targeted properties, see PropCheck.TargetedPBT. {:skip_mfas, [mfa]} When checking a module&#39;s specs, PropEr will not test the specified MFAs. Default is []. Remark: PropCheck does not support spec-testing. {false_positive_mfas, ((mfa(), args::[any], {:fail, result::any} | {:error | :exit | :throw, reason::any}) -&gt; boolean) | :undefined When checking a module&#39;s spec(s), PropEr will treat a counterexample as a false positive if the user supplied function returns true. Otherwise, PropEr will treat the counterexample as it normally does. The inputs to the user supplied function are the MFA, the arguments passed to the MFA, and the result returned from the MFA or an exception with it&#39;s reason. If needed, the user supplied function can call :erlang.get_stacktrace/0. Default is :undefined. Remark: PropCheck does not support spec-testing. PropCheck Specific Options {detect_exceptions | boolean} PropCheck can attempt to detect exceptions thrown or errors raised during a testing function. It will output the error and the corresponding stacktrace. Errors The following errors may be encountered during testing. The term provided for each error is the error type returned by quickcheck/2 in case such an error occurs. Normally, a message is also printed on screen describing the error. :arity_limit The random instance generation subsystem has failed to produce a function of the desired arity. Please recompile PropEr with a suitable value for ?MAX_ARITY (defined in proper_internal.hrl). This error should only be encountered during normal operation. :cant_generate The random instance generation subsystem has failed to produce an instance that satisfies some such_that/2 constraint. You should either increase the :constraint_tries limit, loosen the failing constraint, or make it non-strict. This error should only be encountered during normal operation. :cant_satisfy All the tests were rejected because no produced test case would pass all implies/2 checks. You should loosen the failing implies/2 constraint(s). This error should only be encountered during normal operation. :non_boolean_result The property code returned a non-boolean result. Please fix your property. :rejected Only encountered during re-checking, the counterexample does not match the property, since the counterexample doesn&#39;t pass an implies/2 check. :too_many_instances Only encountered during re-checking, the counterexample does not match the property, since the counterexample contains more instances than there are forall/2s in the property. :type_mismatch The variables&#39; and types&#39; structures inside a forall/2 don&#39;t match. Please check your properties. {:typeserver, sub_error} The typeserver encountered an error. The sub_error field contains specific information regarding the error. {:unexpected, result} A test returned an unexpected result during normal operation. If you ever get this error, it means that you have found a bug in PropEr please send an error report to the maintainers and remember to include both the failing test case and the output of the program, if possible. {:unrecognized_option, option} option is not an option that PropEr understands. Options Available in the Environment PROPCHECK_VERBOSE can be used to enable or disable verbose output globally. PROPCHECK_DETECT_EXCEPTIONS can be used to enable or disable exception and error detection globally. Acknowledgments Very much of the documentation is directly taken from the proper API documentation."},{"ref":"PropCheck.html#aggregate/2","title":"PropCheck.aggregate/2","type":"function","doc":"Same as collect/2, but accepts a list of categories under which to classify the produced test case."},{"ref":"PropCheck.html#aggregate/3","title":"PropCheck.aggregate/3","type":"function","doc":"Same as collect/3, but accepts a list of categories under which to classify the produced test case."},{"ref":"PropCheck.html#check/3","title":"PropCheck.check/3","type":"function","doc":"Re-checks a specific counterexample cexm against the property outer_test that it previously falsified."},{"ref":"PropCheck.html#check_spec/2","title":"PropCheck.check_spec/2","type":"function","doc":"Tests the accuracy of an exported function&#39;s spec."},{"ref":"PropCheck.html#check_specs/2","title":"PropCheck.check_specs/2","type":"function","doc":"Tests all exported, -speced functions of a module mod against their spec."},{"ref":"PropCheck.html#classify/3","title":"PropCheck.classify/3","type":"function","doc":"Same as collect/2, but can accept both a single category and a list of categories. count is a boolean flag: when false, the particular test case will not be counted."},{"ref":"PropCheck.html#collect/2","title":"PropCheck.collect/2","type":"function","doc":"Specifies that test cases produced by this property should be categorized under the term category. This field can be an expression or statement block that evaluates to any term. All produced categories are printed at the end of testing (in case no test fails) along with the percentage of test cases belonging to each category. Multiple collect wrappers are allowed in a single property, in which case the percentages for each collect wrapper are printed separately."},{"ref":"PropCheck.html#collect/3","title":"PropCheck.collect/3","type":"function","doc":"Same as collect/2, but also accepts a fun printer to be used as the stats printer."},{"ref":"PropCheck.html#conjunction/1","title":"PropCheck.conjunction/1","type":"function","doc":"Returns a property that is true only if all of the sub-properties sub_properties are true. Each sub-property should be tagged with a distinct atom. If this property fails, each failing sub-property will be reported and saved inside the counterexample along with its tag."},{"ref":"PropCheck.html#counterexample/0","title":"PropCheck.counterexample/0","type":"function","doc":"Retrieves the last (simplest) counterexample produced by PropCheck during the most recent testing run."},{"ref":"PropCheck.html#counterexample/2","title":"PropCheck.counterexample/2","type":"function","doc":"Equivalent to quickcheck/2, also accepting a list of options."},{"ref":"PropCheck.html#counterexamples/0","title":"PropCheck.counterexamples/0","type":"function","doc":"Returns a counterexample for each failing property of the most recent module testing run."},{"ref":"PropCheck.html#delay/1","title":"PropCheck.delay/1","type":"macro","doc":"Delays the evaluation of expr. Required for defining recursive generators and similar situations."},{"ref":"PropCheck.html#equals/2","title":"PropCheck.equals/2","type":"function","doc":"A custom property that evaluates to true only if a === b, else evaluates to false and prints &quot;A != B&quot; on the screen. Examples iex&gt; use PropCheck iex&gt; quickcheck(equals(:ok, :ok)) :true iex&gt; use PropCheck iex&gt; quickcheck( ...&gt; forall x &lt;- :ok do ...&gt; equals(:ok, x) ...&gt; end) :true iex&gt; use PropCheck iex&gt; quickcheck( ...&gt; forall x &lt;- :not_ok do ...&gt; equals(:ok, x) ...&gt; end) :false"},{"ref":"PropCheck.html#fails/1","title":"PropCheck.fails/1","type":"function","doc":"Specifies that we expect the property property to fail for some input. The property will be considered failing if it passes all the tests."},{"ref":"PropCheck.html#forall/3","title":"PropCheck.forall/3","type":"macro","doc":"A property that should hold for all values generated. iex&gt; use PropCheck iex&gt; quickcheck( ...&gt; forall n &lt;- nat() do ...&gt; n &gt;= 0 ...&gt; end) true If you need more than one generator, collect the generator names and the generators definitions in tuples or lists, respectively: iex&gt; use PropCheck iex&gt; quickcheck( ...&gt; forall [n, l] &lt;- [nat(), list(nat())] do ...&gt; n * Enum.sum(l) &gt;= 0 ...&gt; end ...&gt;) true Similar to let/2, multiple types can also be combined in a list of bindings: iex&gt; use PropCheck iex&gt; quickcheck( ...&gt; forall [n &lt;- nat(), l &lt;- list(nat())] do ...&gt; n * Enum.sum(l) &gt;= 0 ...&gt; end ...&gt;) true forall allows using ExUnit assertions. By default (:quiet), no output is generated if an assertion does not hold: iex&gt; use PropCheck iex&gt; quickcheck( ...&gt; forall n &lt;- nat() do ...&gt; assert n &lt; 0 ...&gt; end ...&gt;) false Use :verbose to enable output on failed ExUnit assertions. Setting Verbosity on the Command Line Apart from setting :verbose or :quiet explicitly in the source code, the PROPCHECK_VERBOSE environment variable can also be used to set the verbosity. PROPCHECK_VERBOSE=1 sets :verbose, while PROPCHECK_VERBOSE=0 sets :quiet."},{"ref":"PropCheck.html#implies/2","title":"PropCheck.implies/2","type":"macro","doc":"A property that is only tested if a condition is true. This wrapper only makes sense when in the scope of at least one forall. The precondition field must be a boolean expression or a statement block that returns a boolean. If the precondition evaluates to false for the variable instances produced in the enclosing forall wrappers, the test case is rejected (it doesn&#39;t count as a failing test case), and PropCheck starts over with a new random test case. Also, in verbose mode, an x is printed on screen. iex&gt; use PropCheck iex&gt; require Integer iex&gt; quickcheck( ...&gt; forall n &lt;- nat() do ...&gt; implies rem(n, 2) == 0, do: Integer.is_even n ...&gt; end ...&gt;) true"},{"ref":"PropCheck.html#is_property/1","title":"PropCheck.is_property/1","type":"macro","doc":""},{"ref":"PropCheck.html#lazy/1","title":"PropCheck.lazy/1","type":"macro","doc":"Mostly internally used macro to create a lazy value for proper. The parameter delayed_value needs to be an already delayed value."},{"ref":"PropCheck.html#let/2","title":"PropCheck.let/2","type":"macro","doc":"Binds a generator to a name for use in another generator. The binding has the generator syntax x &lt;- type. To produce an instance of this type, all appearances of the variables in x are replaced inside generator by their corresponding values in a randomly generated instance of type. It&#39;s OK for the gen part to evaluate to a type - in that case, an instance of the inner type is generated recursively. iex&gt; use PropCheck iex&gt; even = let n &lt;- nat() do ...&gt; n * 2 ...&gt; end iex&gt; quickcheck( ...&gt; forall n &lt;- even do ...&gt; rem(n, 2) == 0 ...&gt; end) true If you require more than one type, put the pairs of variable and type into a list as shown in the example below. iex&gt; use PropCheck iex&gt; even_factor = let [n &lt;- nat(), m &lt;- nat()] do ...&gt; n * m * 2 ...&gt; end iex&gt; quickcheck( ...&gt; forall n &lt;- even_factor do ...&gt; rem(n, 2) == 0 ...&gt; end) true You also can refer to variables declared in the current let scope with ^. Such variables are not allowed to form a dependency cycle. You cannot apply ^ to variables declared outside of the let macro. iex&gt; use PropCheck iex&gt; non_decreasing = ...&gt; let [m &lt;- integer(^l, :inf), l &lt;- integer(), h &lt;- integer(^m, :inf)] do ...&gt; {l, m, h} ...&gt; end iex&gt; quickcheck( ...&gt; forall {l, m, h} &lt;- non_decreasing do ...&gt; l &lt;= m and m &lt;= h ...&gt; end) true Similar to forall/2, multiple types can also be put into tuples or lists: iex&gt; use PropCheck iex&gt; even_factor = let {n, m} &lt;- {nat(), nat()} do ...&gt; n * m * 2 ...&gt; end iex&gt; quickcheck( ...&gt; forall n &lt;- even_factor do ...&gt; rem(n, 2) == 0 ...&gt; end) true"},{"ref":"PropCheck.html#let_shrink/2","title":"PropCheck.let_shrink/2","type":"macro","doc":"A combination of a let and a shrink macro. Instances are generated by applying a randomly generated list of values inside generator (just like a let, with the added constraint that the variables and types must be provided in a list - alternatively, list_of_types may be a list or vector type). When shrinking instances of such a type, the sub-instances that were combined to produce it are first tried in place of the failing instance. One possible use is shown in the tree example. A recursive tree generator with an efficient shrinking: pick each of the subtrees in place of the tree that fails the property. l and r are assigned smaller versions of the tree thus achieving a better (or more appropriate) shrinking. iex&gt; use PropCheck iex&gt; tree_gen = fn (0, _, _) -&gt; :leaf ...&gt; (s, g, tree) -&gt; ...&gt; frequency [ ...&gt; {1, tree.(0, g, tree)}, ...&gt; {9, let_shrink([ ...&gt; l &lt;- tree.(div(s, 2), g, tree), ...&gt; r &lt;- tree.(div(s, 2), g, tree) ...&gt; ]) do ...&gt; {:node, g, l, r} ...&gt; end ...&gt; } ...&gt; ] ...&gt; end iex&gt; tree = fn(g) -&gt; sized(s, tree_gen.(s, g, tree_gen)) end iex&gt; quickcheck( ...&gt; forall t &lt;- tree.(int()) do ...&gt; t == :leaf or is_tuple(t) ...&gt; end ...&gt;) true"},{"ref":"PropCheck.html#measure/3","title":"PropCheck.measure/3","type":"function","doc":"A function that collects numeric statistics on the produced instances. The number (or numbers) provided are collected and some statistics over the collected sample are printed at the end of testing (in case no test fails), prepended with title, which should be an atom or string."},{"ref":"PropCheck.html#module/2","title":"PropCheck.module/2","type":"function","doc":"Tests all properties (i.e., all 0-arity functions whose name begins with prop_) exported from module mod"},{"ref":"PropCheck.html#numtests/2","title":"PropCheck.numtests/2","type":"function","doc":"Specifies the number N of tests to run when testing the property property. Default is 100. This function only changes the number of the tests, but not the size of a test."},{"ref":"PropCheck.html#on_output/2","title":"PropCheck.on_output/2","type":"function","doc":"Specifies an output function print to be used by PropCheck for all output printing during the testing of property property. This wrapper is equivalent to the on_output option."},{"ref":"PropCheck.html#produce/3","title":"PropCheck.produce/3","type":"function","doc":"Sample values from a generator gen. Example iex&gt; use PropCheck iex&gt; produce(1) {:ok, 1} iex&gt; nat() |&gt; produce() |&gt; Kernel.elem(1) |&gt; is_integer() true iex&gt; nat() |&gt; list() |&gt; produce() |&gt; Kernel.elem(1) |&gt; is_list() true"},{"ref":"PropCheck.html#property_setup/2","title":"PropCheck.property_setup/2","type":"macro","doc":"Adds a setup function to the property which will be called before the first test. This function has to return a finalize function of arity 0, which should return the atom :ok, that will be called after the last test. It is possible to use multiple property_setup macros on the same property."},{"ref":"PropCheck.html#quickcheck/1","title":"PropCheck.quickcheck/1","type":"function","doc":"Runs PropEr on the property outer_test."},{"ref":"PropCheck.html#quickcheck/2","title":"PropCheck.quickcheck/2","type":"function","doc":"Same as quickcheck/1, but also accepts a list of options."},{"ref":"PropCheck.html#sample_shrink/2","title":"PropCheck.sample_shrink/2","type":"function","doc":"Generates a random instance of Type, of size Size, then shrinks it as far as it goes. The value produced on each step of the shrinking process is printed on the screen."},{"ref":"PropCheck.html#shrink/2","title":"PropCheck.shrink/2","type":"macro","doc":"Defines the shrinking of a generator. shrink creates a type whose instances are generated by evaluating the statement block generator (this may evaluate to a type, which will then be generated recursively). If an instance of such a type is to be shrunk, the generators in alt_gens are first run to produce hopefully simpler instances of the type. Thus, the generators in the second argument should be simpler than the default. The simplest ones should be at the front of the list, since those are the generators preferred by the shrinking subsystem. Like the main generator, the alternatives may also evaluate to a type, which is generated recursively. iex&gt; use PropCheck iex&gt; quickcheck( ...&gt; forall n &lt;- shrink(pos_integer(), [0]) do ...&gt; rem(n, 2) == 0 ...&gt; end) false"},{"ref":"PropCheck.html#sized/2","title":"PropCheck.sized/2","type":"macro","doc":"Changes the maximum size of the generated instances. sized creates a new type, whose instances are produced by replacing all appearances of the size parameter inside the statement block generator with the value of the size parameter. It&#39;s OK for the generator to return a type - in that case, an instance of the inner type is generated recursively. An example for sized is shown in the documentation of let_shrink/2."},{"ref":"PropCheck.html#such_that/2","title":"PropCheck.such_that/2","type":"macro","doc":"This produces a specialization of a generator, encoded as a binding of form x &lt;- type (as in the let macro). The specialization of members of type that satisfy the constraint condition - that is, those members for which the function fn(x) -&gt; condition end returns true. If the constraint is very strict - that is, only a small percentage of instances of type pass the test - it will take a lot of tries for the instance generation subsystem to randomly produce a valid instance. This will result in slower testing, and testing may even be stopped short, in case the :constraint_tries limit is reached (see the &quot;Options&quot; section). It probably will be more appropriate to generate valid instances of the specialized type using the let macro. Also make sure that even small instances can satisfy the constraint, since PropEr will only increase the size of generated instances when it can find an instance passing the constraint test. If this is not possible, you can instruct PropEr to start at a larger size, by supplying a suitable value for the :start_size option (see the &quot;Options&quot; section). In the second example below, :start_size set to 2 is required, because otherwise PropCheck would first try to generate lists of size 1 (or less), but such a list will never satisfy the length constraint, and after :constraint_tries limit has been reached, PropCheck would give up. Examples iex&gt; use PropCheck iex&gt; even = such_that n &lt;- nat(), when: rem(n, 2) == 0 iex&gt; quickcheck( ...&gt; forall n &lt;- even do ...&gt; rem(n, 2) == 0 ...&gt; end) true iex&gt; use PropCheck iex&gt; twofer_list = such_that l &lt;- list(nat()), when: length(l) &gt; 1 iex&gt; quickcheck( ...&gt; forall [_ | t] &lt;- twofer_list do ...&gt; Enum.any?(t) ...&gt; end, [start_size: 2]) true"},{"ref":"PropCheck.html#such_that_maybe/2","title":"PropCheck.such_that_maybe/2","type":"macro","doc":"Equivalent to the such_that macro, but the constraint condition is considered non-strict: if the :constraint_tries limit is reached, the generator will just return an instance of type instead of failing, even if that instance doesn&#39;t satisfy the constraint. iex&gt; use PropCheck iex&gt; even = such_that_maybe n &lt;- nat(), when: rem(n, 2) == 0 iex&gt; quickcheck( ...&gt; forall n &lt;- even do ...&gt; rem(n, 2) == 0 ...&gt; end) true"},{"ref":"PropCheck.html#timeout/2","title":"PropCheck.timeout/2","type":"macro","doc":"Signifies that prop should be considered failing if it takes more than time_limit milliseconds to return. The purpose of this wrapper is to test code that may hang if something goes wrong. timeout cannot contain any more wrappers. iex&gt; use PropCheck iex&gt; quickcheck( ...&gt; timeout(100, forall n &lt;- nat() do ...&gt; :ok == :timer.sleep(n*100) ...&gt; end) ...&gt; ) false"},{"ref":"PropCheck.html#trap_exit/1","title":"PropCheck.trap_exit/1","type":"macro","doc":"If the code inside prop spawns and links to a process that dies abnormally, PropEr will catch the exit signal and treat it as a test failure, instead of crashing. trap_exit cannot contain any more wrappers. iex&gt; use PropCheck iex&gt; quickcheck( ...&gt; trap_exit(forall n &lt;- nat() do ...&gt; # this must fail ...&gt; pid = spawn_link(fn() -&gt; n / 0 end) ...&gt; # wait for arrivial of the dieing linked process signal ...&gt; :timer.sleep(50) ...&gt; true # ...&gt; end) ...&gt; ) false"},{"ref":"PropCheck.html#when_fail/2","title":"PropCheck.when_fail/2","type":"macro","doc":"Execute an action, if the property fails. The action field should contain an expression or statement block that produces some side-effect (e.g. prints something to the screen). In case this test fails, action will be executed. Note that the output of such actions is not affected by the verbosity setting of the main application. iex&gt; use PropCheck iex&gt; quickcheck( ...&gt; when_fail(false, IO.puts &quot;when_fail: Property failed&quot;) ...&gt;) false"},{"ref":"PropCheck.html#with_title/1","title":"PropCheck.with_title/1","type":"function","doc":"A predefined function that accepts an atom or string and returns a stats printing function which is equivalent to the default one, but prints the given title title above the statistics."},{"ref":"PropCheck.html#t:counterexample/0","title":"PropCheck.counterexample/0","type":"type","doc":""},{"ref":"PropCheck.html#t:error/0","title":"PropCheck.error/0","type":"type","doc":""},{"ref":"PropCheck.html#t:error_reason/0","title":"PropCheck.error_reason/0","type":"type","doc":""},{"ref":"PropCheck.html#t:false_positive_mfas/0","title":"PropCheck.false_positive_mfas/0","type":"type","doc":""},{"ref":"PropCheck.html#t:long_module_result/0","title":"PropCheck.long_module_result/0","type":"type","doc":""},{"ref":"PropCheck.html#t:long_result/0","title":"PropCheck.long_result/0","type":"type","doc":""},{"ref":"PropCheck.html#t:module_result/0","title":"PropCheck.module_result/0","type":"type","doc":""},{"ref":"PropCheck.html#t:outer_test/0","title":"PropCheck.outer_test/0","type":"type","doc":""},{"ref":"PropCheck.html#t:output_fun/0","title":"PropCheck.output_fun/0","type":"type","doc":""},{"ref":"PropCheck.html#t:result/0","title":"PropCheck.result/0","type":"type","doc":""},{"ref":"PropCheck.html#t:sample/0","title":"PropCheck.sample/0","type":"type","doc":""},{"ref":"PropCheck.html#t:short_module_result/0","title":"PropCheck.short_module_result/0","type":"type","doc":""},{"ref":"PropCheck.html#t:short_result/0","title":"PropCheck.short_result/0","type":"type","doc":""},{"ref":"PropCheck.html#t:size/0","title":"PropCheck.size/0","type":"type","doc":""},{"ref":"PropCheck.html#t:stats_printer/0","title":"PropCheck.stats_printer/0","type":"type","doc":""},{"ref":"PropCheck.html#t:test/0","title":"PropCheck.test/0","type":"type","doc":""},{"ref":"PropCheck.html#t:title/0","title":"PropCheck.title/0","type":"type","doc":""},{"ref":"PropCheck.html#t:type/0","title":"PropCheck.type/0","type":"type","doc":""},{"ref":"PropCheck.html#t:user_opt/0","title":"PropCheck.user_opt/0","type":"type","doc":""},{"ref":"PropCheck.html#t:user_opts/0","title":"PropCheck.user_opts/0","type":"type","doc":""},{"ref":"PropCheck.BasicTypes.html","title":"PropCheck.BasicTypes","type":"module","doc":"This modules contains all basic type generators from PropEr. It is automatically available by use PropCheck. Acknowledgments The functions defined here are delegated to the corresponding definition proper_types. Also most of the documentation is copied over from there."},{"ref":"PropCheck.BasicTypes.html#any/0","title":"PropCheck.BasicTypes.any/0","type":"function","doc":"All Elixir terms (that PropEr can produce). For reasons of efficiency, functions are never produced as instances of this type. CAUTION: Instances of this type are expensive to produce, shrink and instance- check, both in terms of processing time and consumed memory. Only use this type if you are certain that you need it."},{"ref":"PropCheck.BasicTypes.html#arity/0","title":"PropCheck.BasicTypes.arity/0","type":"function","doc":"Arity is a byte value, i.e. integer(0, 255)"},{"ref":"PropCheck.BasicTypes.html#atom/0","title":"PropCheck.BasicTypes.atom/0","type":"function","doc":"All atoms. All atoms used internally by PropEr start with a :$ , so such atoms will never be produced as instances of this type. You should also refrain from using such atoms in your code, to avoid a potential clash. Instances shrink towards the empty atom, :&quot;&quot;."},{"ref":"PropCheck.BasicTypes.html#binary/0","title":"PropCheck.BasicTypes.binary/0","type":"function","doc":"All binaries. Instances shrink towards the empty binary, &quot;&quot;."},{"ref":"PropCheck.BasicTypes.html#binary/1","title":"PropCheck.BasicTypes.binary/1","type":"function","doc":"All binaries with a byte size of length. length must be an Elixir expression that evaluates to a non-negative integer. Instances shrink towards binaries of zeroes."},{"ref":"PropCheck.BasicTypes.html#bitstring/0","title":"PropCheck.BasicTypes.bitstring/0","type":"function","doc":"All bitstrings. Instances shrink towards the empty bitstring, &quot;&quot;."},{"ref":"PropCheck.BasicTypes.html#bitstring/1","title":"PropCheck.BasicTypes.bitstring/1","type":"function","doc":"All bitstrings with a bit size of length. length must be an Elixir expression that evaluates to a non-negative integer. Instances shrink towards bitstrings of zeroes."},{"ref":"PropCheck.BasicTypes.html#bool/0","title":"PropCheck.BasicTypes.bool/0","type":"function","doc":"bool is equivalent to boolean"},{"ref":"PropCheck.BasicTypes.html#boolean/0","title":"PropCheck.BasicTypes.boolean/0","type":"function","doc":"The atoms true and false. Instances shrink towards false."},{"ref":"PropCheck.BasicTypes.html#byte/0","title":"PropCheck.BasicTypes.byte/0","type":"function","doc":"Byte values, i.e. integer(0, 255)"},{"ref":"PropCheck.BasicTypes.html#char/0","title":"PropCheck.BasicTypes.char/0","type":"function","doc":"Char values (16 bit for some reason), i.e. integer(0, 0xffff)"},{"ref":"PropCheck.BasicTypes.html#char_list/0","title":"PropCheck.BasicTypes.char_list/0","type":"function","doc":"An Erlang string, i.e. list(char)"},{"ref":"PropCheck.BasicTypes.html#choose/2","title":"PropCheck.BasicTypes.choose/2","type":"function","doc":"choose is equivalent to integer(low, high)"},{"ref":"PropCheck.BasicTypes.html#default/2","title":"PropCheck.BasicTypes.default/2","type":"function","doc":"Adds a default value, default_value, to type. The default serves as a primary shrinking target for instances, while it is also chosen by the random instance generation subsystem half the time."},{"ref":"PropCheck.BasicTypes.html#elements/1","title":"PropCheck.BasicTypes.elements/1","type":"function","doc":"elements is equivalent to union([..])"},{"ref":"PropCheck.BasicTypes.html#exactly/1","title":"PropCheck.BasicTypes.exactly/1","type":"function","doc":"Singleton type consisting only of value. value must be an evaluated term. Also written simply as value."},{"ref":"PropCheck.BasicTypes.html#fixed_list/1","title":"PropCheck.BasicTypes.fixed_list/1","type":"function","doc":"All lists whose i-th element is an instance of the type at index i of list_of_types. Also written simply as a list of types."},{"ref":"PropCheck.BasicTypes.html#float/0","title":"PropCheck.BasicTypes.float/0","type":"function","doc":"All floats, i.e. float(:inf, :inf)"},{"ref":"PropCheck.BasicTypes.html#float/2","title":"PropCheck.BasicTypes.float/2","type":"function","doc":"All floats between low and high, bounds included. low and high must be Elixir expressions that evaluate to floats, with Low =&lt; high. Additionally, low and high may have the value :inf, in which case they represent minus infinity and plus infinity respectively. Instances shrink towards 0.0 if low =&lt; 0.0 =&lt; high, or towards the bound with the smallest absolute value otherwise."},{"ref":"PropCheck.BasicTypes.html#frequency/1","title":"PropCheck.BasicTypes.frequency/1","type":"function","doc":"frequency is equivalent to weighted_union([..])"},{"ref":"PropCheck.BasicTypes.html#function/2","title":"PropCheck.BasicTypes.function/2","type":"function","doc":"All pure functions that map instances of arg_types to instances of ret_type. The syntax function(arity, ret_type) is also acceptable."},{"ref":"PropCheck.BasicTypes.html#function0/1","title":"PropCheck.BasicTypes.function0/1","type":"function","doc":"A function with 0 parameters, i.e. function(0, ret_type)"},{"ref":"PropCheck.BasicTypes.html#function1/1","title":"PropCheck.BasicTypes.function1/1","type":"function","doc":"A function with 1 parameter, i.e. function(1, ret_type)"},{"ref":"PropCheck.BasicTypes.html#function2/1","title":"PropCheck.BasicTypes.function2/1","type":"function","doc":"A function with 2 parameters, i.e. function(2, ret_type)"},{"ref":"PropCheck.BasicTypes.html#function3/1","title":"PropCheck.BasicTypes.function3/1","type":"function","doc":"A function with 3 parameters, i.e. function(3, ret_type)"},{"ref":"PropCheck.BasicTypes.html#function4/1","title":"PropCheck.BasicTypes.function4/1","type":"function","doc":"A function with 4 parameters, i.e. function(4, ret_type)"},{"ref":"PropCheck.BasicTypes.html#int/0","title":"PropCheck.BasicTypes.int/0","type":"function","doc":"Small integers (bound by the current value of the size parameter). Instances shrink towards 0."},{"ref":"PropCheck.BasicTypes.html#integer/0","title":"PropCheck.BasicTypes.integer/0","type":"function","doc":"All integers, i.e. integer(:inf, :inf)"},{"ref":"PropCheck.BasicTypes.html#integer/2","title":"PropCheck.BasicTypes.integer/2","type":"function","doc":"All integers between low and high, bounds included. low and high must be Elixir expressions that evaluate to integers, with low =&lt; high. Additionally, low and high may have the value :inf, in which case they represent minus infinity and plus infinity respectively. Instances shrink towards 0 if low =&lt; 0 =&lt; high, or towards the bound with the smallest absolute value otherwise."},{"ref":"PropCheck.BasicTypes.html#large_int/0","title":"PropCheck.BasicTypes.large_int/0","type":"function","doc":"Large_int is equivalent to integer"},{"ref":"PropCheck.BasicTypes.html#list/0","title":"PropCheck.BasicTypes.list/0","type":"function","doc":"List of any types, i.e. list(any)"},{"ref":"PropCheck.BasicTypes.html#list/1","title":"PropCheck.BasicTypes.list/1","type":"function","doc":"All lists containing elements of type elem_type. Instances shrink towards the empty list, []."},{"ref":"PropCheck.BasicTypes.html#loose_tuple/1","title":"PropCheck.BasicTypes.loose_tuple/1","type":"function","doc":"Tuples whose elements are all of type elem_type. Instances shrink towards the 0-size tuple, {}."},{"ref":"PropCheck.BasicTypes.html#map/2","title":"PropCheck.BasicTypes.map/2","type":"function","doc":"A map whose keys are defined by the generator k and values by the generator v."},{"ref":"PropCheck.BasicTypes.html#nat/0","title":"PropCheck.BasicTypes.nat/0","type":"function","doc":"Small Small non-negative integers (bound by the current value of the size parameter). Instances shrink towards 0."},{"ref":"PropCheck.BasicTypes.html#neg_integer/0","title":"PropCheck.BasicTypes.neg_integer/0","type":"function","doc":"Negative integers, i.e. integer(:inf, -1)"},{"ref":"PropCheck.BasicTypes.html#non_empty/1","title":"PropCheck.BasicTypes.non_empty/1","type":"function","doc":"This is a predefined constraint that can be applied to random-length list and binary types to ensure that the produced values are never empty. Use for e.g. list/0, char_list/0, binary/0"},{"ref":"PropCheck.BasicTypes.html#non_neg_float/0","title":"PropCheck.BasicTypes.non_neg_float/0","type":"function","doc":"Non negative floats, i.e. float(0.0, inf)"},{"ref":"PropCheck.BasicTypes.html#non_neg_integer/0","title":"PropCheck.BasicTypes.non_neg_integer/0","type":"function","doc":"Non negative integers, i.e. integer(0, :inf)"},{"ref":"PropCheck.BasicTypes.html#noshrink/1","title":"PropCheck.BasicTypes.noshrink/1","type":"function","doc":"Creates a new type which is equivalent to type, but whose instances are never shrunk by the shrinking subsystem."},{"ref":"PropCheck.BasicTypes.html#number/0","title":"PropCheck.BasicTypes.number/0","type":"function","doc":"Numbers are integers or floats, i.e. union([integer(), float()])"},{"ref":"PropCheck.BasicTypes.html#oneof/1","title":"PropCheck.BasicTypes.oneof/1","type":"function","doc":"oneof is equivalent to union([..])"},{"ref":"PropCheck.BasicTypes.html#ordered_list/1","title":"PropCheck.BasicTypes.ordered_list/1","type":"function","doc":"All sorted lists containing elements of type elem_type. Instances shrink towards the empty list, []."},{"ref":"PropCheck.BasicTypes.html#parameter/1","title":"PropCheck.BasicTypes.parameter/1","type":"function","doc":"Returns the value associated with parameter, or :undefined in case parameter is not associated with any value. Association occurs with calling with_parameter/3 or with_parameters/2 before."},{"ref":"PropCheck.BasicTypes.html#parameter/2","title":"PropCheck.BasicTypes.parameter/2","type":"function","doc":"Returns the value associated with parameter, or default in case parameter is not associated with any value. Association occurs with calling with_parameter/3 or with_parameters/2 before."},{"ref":"PropCheck.BasicTypes.html#pos_integer/0","title":"PropCheck.BasicTypes.pos_integer/0","type":"function","doc":"Strictly positive integers, i.e. integer(1, :inf)"},{"ref":"PropCheck.BasicTypes.html#range/2","title":"PropCheck.BasicTypes.range/2","type":"function","doc":"A range is equivalent to integers"},{"ref":"PropCheck.BasicTypes.html#real/0","title":"PropCheck.BasicTypes.real/0","type":"function","doc":"real is equivalent to float"},{"ref":"PropCheck.BasicTypes.html#resize/2","title":"PropCheck.BasicTypes.resize/2","type":"function","doc":"Overrides the size parameter used when generating instances of type with new_size. Has no effect on size-less types, such as unions. Also, this will not affect the generation of any internal types contained in type, such as the elements of a list - those will still be generated using the test-wide value of size. One use of this function is to modify types to produce instances that grow faster or slower, like so: iex&gt; quickcheck(forall l &lt;- list(integer()) do ...&gt; length(l) &lt;= 42 ...&gt; end) true iex&gt; long_list = sized(size, resize(size * 2, list(integer()))) iex&gt; really_long = such_that_maybe l &lt;- long_list, when: ...&gt; length(l) &gt; 42 iex&gt; quickcheck(forall l &lt;- really_long do ...&gt; (length(l) &lt;= 84) ...&gt; |&gt; measure(&quot;List length&quot;, length l) ...&gt; |&gt; collect(length l) ...&gt; end) true The above specifies a list type that grows twice as fast as normal lists."},{"ref":"PropCheck.BasicTypes.html#return/1","title":"PropCheck.BasicTypes.return/1","type":"function","doc":"return is equivalent to exactly"},{"ref":"PropCheck.BasicTypes.html#shrink_list/1","title":"PropCheck.BasicTypes.shrink_list/1","type":"function","doc":"A type that generates exactly the list list. Instances shrink towards shorter sublists of the original list."},{"ref":"PropCheck.BasicTypes.html#term/0","title":"PropCheck.BasicTypes.term/0","type":"function","doc":"Term is a synonym for any"},{"ref":"PropCheck.BasicTypes.html#timeout/0","title":"PropCheck.BasicTypes.timeout/0","type":"function","doc":"timeout values, i.e. union([non_neg_integer() | :infinity])"},{"ref":"PropCheck.BasicTypes.html#tuple/0","title":"PropCheck.BasicTypes.tuple/0","type":"function","doc":"Tuples of any types, i.e. loose_tuple(any)"},{"ref":"PropCheck.BasicTypes.html#tuple/1","title":"PropCheck.BasicTypes.tuple/1","type":"function","doc":"All tuples whose i-th element is an instance of the type at index i of list_of_types. Also written simply as a tuple of types."},{"ref":"PropCheck.BasicTypes.html#union/1","title":"PropCheck.BasicTypes.union/1","type":"function","doc":"The union of all types in list_of_types. list_of_types can&#39;t be empty. The random instance generator is equally likely to choose any one of the types in list_of_types. The shrinking subsystem will always try to shrink an instance of a type union to an instance of the first type in list_of_types, thus you should write the simplest case first."},{"ref":"PropCheck.BasicTypes.html#utf8/0","title":"PropCheck.BasicTypes.utf8/0","type":"function","doc":"utf8-encoded unbounded size binary"},{"ref":"PropCheck.BasicTypes.html#utf8/1","title":"PropCheck.BasicTypes.utf8/1","type":"function","doc":"utf8-encoded bounded upper size binary."},{"ref":"PropCheck.BasicTypes.html#utf8/2","title":"PropCheck.BasicTypes.utf8/2","type":"function","doc":"Bounded upper size utf8 binary, codepoint length =&lt; MaxCodePointSize. Limiting codepoint size can be useful when applications do not accept full unicode range. For example, MySQL in utf8 encoding accepts only 3-byte unicode codepoints in VARCHAR fields. If unbounded length is needed, use :inf as first argument."},{"ref":"PropCheck.BasicTypes.html#vector/2","title":"PropCheck.BasicTypes.vector/2","type":"function","doc":"All lists of length length containing elements of type elem_type. length must be an Elixir expression that evaluates to a non-negative integer."},{"ref":"PropCheck.BasicTypes.html#weighted_default/2","title":"PropCheck.BasicTypes.weighted_default/2","type":"function","doc":"A specialization of default/2. Parameters default and type are assigned weights to be considered by the random instance generator. The shrinking subsystem will ignore the weights and try to shrink using the default value."},{"ref":"PropCheck.BasicTypes.html#weighted_union/1","title":"PropCheck.BasicTypes.weighted_union/1","type":"function","doc":"A specialization of union/1, where each type in list_of_types is assigned a frequency. Frequencies must be Elixir expressions that evaluate to positive integers. Types with larger frequencies are more likely to be chosen by the random instance generator. The shrinking subsystem will ignore the frequencies and try to shrink towards the first type in the list."},{"ref":"PropCheck.BasicTypes.html#with_parameter/3","title":"PropCheck.BasicTypes.with_parameter/3","type":"function","doc":"Associates the atom key parameter with the value value while generating instances of type."},{"ref":"PropCheck.BasicTypes.html#with_parameters/2","title":"PropCheck.BasicTypes.with_parameters/2","type":"function","doc":"Similar to with_parameter/3, but accepts a list of {parameter, value} pairs."},{"ref":"PropCheck.BasicTypes.html#wunion/1","title":"PropCheck.BasicTypes.wunion/1","type":"function","doc":"weighted_union(FreqChoices)"},{"ref":"PropCheck.BasicTypes.html#t:ext_float/0","title":"PropCheck.BasicTypes.ext_float/0","type":"type","doc":"Floats extend by infinity"},{"ref":"PropCheck.BasicTypes.html#t:ext_int/0","title":"PropCheck.BasicTypes.ext_int/0","type":"type","doc":"Integers extend by infinity"},{"ref":"PropCheck.BasicTypes.html#t:ext_non_neg_integer/0","title":"PropCheck.BasicTypes.ext_non_neg_integer/0","type":"type","doc":"Non negative integers extend by infinity"},{"ref":"PropCheck.BasicTypes.html#t:frequency/0","title":"PropCheck.BasicTypes.frequency/0","type":"type","doc":""},{"ref":"PropCheck.BasicTypes.html#t:raw_type/0","title":"PropCheck.BasicTypes.raw_type/0","type":"type","doc":"The internal representation of a basic type in PropEr"},{"ref":"PropCheck.BasicTypes.html#t:size/0","title":"PropCheck.BasicTypes.size/0","type":"type","doc":""},{"ref":"PropCheck.BasicTypes.html#t:type/0","title":"PropCheck.BasicTypes.type/0","type":"type","doc":"The internal representation of a type in PropEr"},{"ref":"PropCheck.BasicTypes.html#t:value/0","title":"PropCheck.BasicTypes.value/0","type":"type","doc":""},{"ref":"PropCheck.FSM.html","title":"PropCheck.FSM","type":"behaviour","doc":"The finite state machine approach for stateful systems, which is closer to Erlang&#39;s gen_fsm model. This module defines the proper_fsm behaviour, useful for testing systems that can be modeled as finite state machines. That is, a finite collection of named states and transitions between them. PropCheck.FSM is closely related to PropCheck.StateM and is, in fact, implemented in terms of that. Test cases generated using PropCheck.FSM will be on precisely the same form as test cases generated using PropCheck.StateM. The difference lies in the way the callback modules are specified. The relation between PropCheck.StateM and PropCheck.FSM is similar to the one between gen_server and gen_fsm in OTP libraries. Due to name conflicts with functions automatically imported from PropCheck.StateM, a fully qualified call is needed in order to use the API functions of PropCheck.FSM. The states of the finite state machine Following the convention used in gen_fsm behaviour, the state is separated into types state_name/0 and some state_data/0. state_name is used to denote a state of the finite state machine and state_data is any relevant information that has to be stored in the model state. States are fully represented as tuples {state_name, state_data}. state_name is usually an atom (i.e. the name of the state), but can also be a tuple. In the latter case, the first element of the tuple must be an atom specifying the name of the state, whereas the rest of the elements can be arbitrary terms specifying state attributes. For example, when implementing the fsm of an elevator which can reach n different floors, the state_name for each floor could be {:floor, k}, 1 &lt;= k &lt;= n. state_data can be an arbitrary term, but is usually a record. Transitions between states A transition transition/0 is represented as a tuple {target_state, {:call, m, f, a}}. This means that performing the specified symbolic call at the current state of the fsm will lead to target_state. The atom :history can be used as target_state to denote that a transition does not change the current state of the fsm. The callback functions The following functions must be exported from the callback module implementing the finite state machine: initial_state/0 initial_data/0 precondition/4 postcondition/5 next_state_data/5 weight/3 In addition to these functions, we also need functions for each state: state_name(s::state_data) ::[transition] There should be one instance of this function for each reachable state state_name of the finite state machine. In case state_name is a tuple the function takes a different form, described just below. The function returns a list of possible transitions (transition/0 ) from the current state. At command generation time, the instance of this function with the same name as the current state&#39;s name is called to return the list of possible transitions. Then, PropEr will randomly choose a transition and, according to that, generate the next symbolic call to be included in the command sequence. However, before the call is actually included, a precondition that might impose constraints on state_data is checked. Note also that PropEr detects transitions that would raise an exception of class &lt;error&gt; at generation time (not earlier) and does not choose them. This feature can be used to include conditional transitions that depend on the state_data/0. state_name(attr1::any, ..., attrN::any, s::type state_data) :: [transition] There should be one instance of this function for each reachable state {state_name, attr1, ..., attrN} of the finite state machine. The function has similar behaviour to state_name/1, described above. The property used This is an example of a property that can be used to test a finite state machine specification. It expects a cleanup function that takes care of removing all artifacts created during tests to enable a clean start for each test case execution. property &quot;fsm&quot; do forall cmds &lt;- commands(__MODULE__) do {_history, _state, result} = run_commands(__MODULE__, cmds) cleanup() result == :ok end end"},{"ref":"PropCheck.FSM.html#command_names/1","title":"PropCheck.FSM.command_names/1","type":"function","doc":"See PropCheck.StateM.command_names/1."},{"ref":"PropCheck.FSM.html#commands/1","title":"PropCheck.FSM.commands/1","type":"function","doc":"A special PropEr type which generates random command sequences, according to a finite state machine specification. The function takes as input the name of a callback module, which contains the fsm specification. The initial state is computed by {mod.initial_state/0, mod:initial_state_data/0}."},{"ref":"PropCheck.FSM.html#commands/2","title":"PropCheck.FSM.commands/2","type":"function","doc":"Similar to commands/1, but generated command sequences always start at a given state. In this case, the first command is always {:init, initial_state = {name, data}} and is used to correctly initialize the state every time the command sequence is run (i.e. during normal execution, while shrinking and when checking a counterexample)."},{"ref":"PropCheck.FSM.html#c:initial_data/0","title":"PropCheck.FSM.initial_data/0","type":"callback","doc":"Specifies what the state data should initially contain. Its result should be deterministic."},{"ref":"PropCheck.FSM.html#c:initial_state/0","title":"PropCheck.FSM.initial_state/0","type":"callback","doc":"Specifies the initial state of the finite state machine. As with PropCheck.StateM.initial_state/0, its result should be deterministic."},{"ref":"PropCheck.FSM.html#more_commands/2","title":"PropCheck.FSM.more_commands/2","type":"function","doc":"See PropCheck.StateM.more_commands/2."},{"ref":"PropCheck.FSM.html#c:next_state_data/5","title":"PropCheck.FSM.next_state_data/5","type":"callback","doc":"Similar to PropCheck.StateM.next_state/3. Specifies how the transition from from to target triggered by call affects the state_data. res refers to the result of call and can be either symbolic or dynamic."},{"ref":"PropCheck.FSM.html#c:postcondition/5","title":"PropCheck.FSM.postcondition/5","type":"callback","doc":"Similar to PropCheck.StateM.postcondition/3. Specifies the postcondition that should hold about the result res of the evaluation of call."},{"ref":"PropCheck.FSM.html#c:precondition/4","title":"PropCheck.FSM.precondition/4","type":"callback","doc":"Similar to PropCheck.StateM.precondition/2. Specifies the precondition that should hold about state_data so that call can be included in the command sequence. In case precondition doesn&#39;t hold, a new transition is chosen using the appropriate state_name/1 generator. It is possible for more than one transitions to be triggered by the same symbolic call and lead to different target states. In this case, at most one of the target states may have a true precondition. Otherwise, PropEr will not be able to detect which transition was chosen and an exception will be raised."},{"ref":"PropCheck.FSM.html#run_commands/2","title":"PropCheck.FSM.run_commands/2","type":"function","doc":"Evaluates a given symbolic command sequence cmds according to the finite state machine specified in mod. The result is a triple of the form {history, fsm_state, result}, similar to PropCheck.StateM.run_commands/2."},{"ref":"PropCheck.FSM.html#run_commands/3","title":"PropCheck.FSM.run_commands/3","type":"function","doc":"Similar to run_commands/2, but also accepts an environment used for symbolic variable evaluation, exactly as described in PropCheck.StateM.run_commands/3."},{"ref":"PropCheck.FSM.html#state_names/1","title":"PropCheck.FSM.state_names/1","type":"function","doc":"Extracts the names of the states from a given command execution history. It is useful in combination with functions such as PropCheck.aggregate/2 in order to collect statistics about state transitions during command execution."},{"ref":"PropCheck.FSM.html#c:weight/3","title":"PropCheck.FSM.weight/3","type":"callback","doc":"This is an optional callback. When it is not defined (or not exported), transitions are chosen with equal probability. When it is defined, it assigns an integer weight to transitions from from to target triggered by symbolic call call. In this case, each transition is chosen with probability proportional to the weight assigned."},{"ref":"PropCheck.FSM.html#t:cmd_result/0","title":"PropCheck.FSM.cmd_result/0","type":"type","doc":""},{"ref":"PropCheck.FSM.html#t:command/0","title":"PropCheck.FSM.command/0","type":"type","doc":""},{"ref":"PropCheck.FSM.html#t:command_list/0","title":"PropCheck.FSM.command_list/0","type":"type","doc":""},{"ref":"PropCheck.FSM.html#t:fsm_result/0","title":"PropCheck.FSM.fsm_result/0","type":"type","doc":""},{"ref":"PropCheck.FSM.html#t:fsm_state/0","title":"PropCheck.FSM.fsm_state/0","type":"type","doc":""},{"ref":"PropCheck.FSM.html#t:history/0","title":"PropCheck.FSM.history/0","type":"type","doc":""},{"ref":"PropCheck.FSM.html#t:mod_name/0","title":"PropCheck.FSM.mod_name/0","type":"type","doc":""},{"ref":"PropCheck.FSM.html#t:result/0","title":"PropCheck.FSM.result/0","type":"type","doc":""},{"ref":"PropCheck.FSM.html#t:state_data/0","title":"PropCheck.FSM.state_data/0","type":"type","doc":""},{"ref":"PropCheck.FSM.html#t:state_name/0","title":"PropCheck.FSM.state_name/0","type":"type","doc":""},{"ref":"PropCheck.FSM.html#t:symbolic_call/0","title":"PropCheck.FSM.symbolic_call/0","type":"type","doc":""},{"ref":"PropCheck.FSM.html#t:symbolic_var/0","title":"PropCheck.FSM.symbolic_var/0","type":"type","doc":""},{"ref":"PropCheck.FSM.html#t:transition/0","title":"PropCheck.FSM.transition/0","type":"type","doc":""},{"ref":"PropCheck.Instrument.html","title":"PropCheck.Instrument","type":"behaviour","doc":"Provides functions and macros for instrument byte code with additional yields and other constructs to ease testing of concurrent programs and state machines. Why is instrumentation important? The Erlang scheduler is relatively predictable and stable with regard to pre-emptive scheduling. This means that every run has more or the less the same amount of virtual machine instructions before a switch to another process happens. These process switches are required to reveal any concurrency bugs. A simple way to provoke more process switches are calls to :erlang.yield() which gives the scheduler the possibility to switch early on to another process. It is not defined if the scheduler reacts on this hint, but it often does and allows for more unpredictable schedules revealing more concurrency bugs. The usual advice is to sprinkle the code under test with manually added calls to :erlang.yield(), but this is a daunting task. Additionally, you need to remove this additional code before production use. The instrumentation The functions in this module automate the instrumentation immediately before running the tests. We instrument call to &quot;interesting&quot; functions of the Erlang and Elixir ecosystem, e.g. calls to GenServer or ets tables. We do this by examining the byte code, checking each function call, and if we found some interesting call target, we add a call to :erlang.yield() immediately before. This is what the PropCheck.YieldInstrumenter module provides. It implements the behaviour Instrument, which requires the implementation of two callbacks handle_function_call/1 and is_instrumentable_function/2. After instrumentation, the code reloading mechanism of the Erlang VM enables the new code and the tests can run. Typical usage To ensure instrumentation before running the tests, you implement the setup_all macro of ExUnit: setup_all do Instrument.instrument_module(Cache, YieldInstrumenter) :ok # no update of a context end In this example, we instrument only a specific module. You can also instrument all modules of an application by calling Instrument.instrument_app(:my_app_under_test, YieldInstrumenter). Implementing your own instrumenter For implementing your own instrumenter, you need to get acquainted with the Erlang Abstract Form (EAF), which is the internal abstract syntax tree available to the Erlang VM at runtime. This format is quite different from the Elixir AST, in particular it has not the regular form but consists of many different structures. This requires a lot of cases to be handled for analyizing the AST. Little helpers for encoding the instrumented code is provided by encode_call/1 and encode_value/1 as well as by prepend_call/2. For debugging and revealing the structure of a specific EAF, you can use print_fun/1."},{"ref":"PropCheck.Instrument.html#call_yield/0","title":"PropCheck.Instrument.call_yield/0","type":"function","doc":"Encodes a call to :erlang.yield() as Erlang Astract Form."},{"ref":"PropCheck.Instrument.html#compile_module/3","title":"PropCheck.Instrument.compile_module/3","type":"function","doc":"Compiles the abstract code of a module and loads it immediately into the VM."},{"ref":"PropCheck.Instrument.html#encode_call/1","title":"PropCheck.Instrument.encode_call/1","type":"function","doc":"Enocdes a call given as tuple {m, f, a} as Erlang Abstract Form."},{"ref":"PropCheck.Instrument.html#encode_call/3","title":"PropCheck.Instrument.encode_call/3","type":"function","doc":"Encodes a call to m.f.(a) as Erlang Abstract Form."},{"ref":"PropCheck.Instrument.html#encode_value/1","title":"PropCheck.Instrument.encode_value/1","type":"function","doc":"Encodes a value as Erlang Astract Form."},{"ref":"PropCheck.Instrument.html#get_forms_of_module/1","title":"PropCheck.Instrument.get_forms_of_module/1","type":"function","doc":"Retrieves the abstract code, i.e. the list of forms, of the given module as found in the code server."},{"ref":"PropCheck.Instrument.html#c:handle_function_call/1","title":"PropCheck.Instrument.handle_function_call/1","type":"callback","doc":"Handle the instrumentation of a (remote) function call. Must return a valid expression in Erlang Abstract Form."},{"ref":"PropCheck.Instrument.html#instrument_app/2","title":"PropCheck.Instrument.instrument_app/2","type":"function","doc":"Instruments all modules of an entire OTP application."},{"ref":"PropCheck.Instrument.html#instrument_module/2","title":"PropCheck.Instrument.instrument_module/2","type":"function","doc":"Takes the object code of the module, instruments it and update the module in the code server with instrumented byte code."},{"ref":"PropCheck.Instrument.html#instrumentable_function/2","title":"PropCheck.Instrument.instrumentable_function/2","type":"function","doc":"Checks if the given function is a candidate for instrumentation, i.e. does something interesting with respect to concurrency. Examples are process handling, handling of shared state or sending and receiving messages."},{"ref":"PropCheck.Instrument.html#c:is_instrumentable_function/2","title":"PropCheck.Instrument.is_instrumentable_function/2","type":"callback","doc":"A callback to decide if the function mod:fun with any arity is a candidate for instrumentation. The default implementation is simply calling instrumentable_function/2."},{"ref":"PropCheck.Instrument.html#is_instrumented?/1","title":"PropCheck.Instrument.is_instrumented?/1","type":"function","doc":"Checks if the code is already instrumented. If not, returns false otherwise returns true"},{"ref":"PropCheck.Instrument.html#prepend_call/2","title":"PropCheck.Instrument.prepend_call/2","type":"function","doc":"Prepends the call to to_be_wrapped_call by a call to new_call. The result of new_call is ignored. All arugments and return values are in Erlang Astract Form."},{"ref":"PropCheck.Instrument.html#print_fun/2","title":"PropCheck.Instrument.print_fun/2","type":"function","doc":"Debugging aid for analyzing code generations. Prints the restructered Erlang code of function fun in module mod. We use Erlang code here, because Elixir source code cannot generated from the byte code format due to macros, which change the compilation process too heavily."},{"ref":"PropCheck.Instrument.html#t:erl_ast_atom_type/0","title":"PropCheck.Instrument.erl_ast_atom_type/0","type":"type","doc":"The type for a node in the Erlang Abstract Form encoding an atom value"},{"ref":"PropCheck.Instrument.html#t:erl_ast_block/0","title":"PropCheck.Instrument.erl_ast_block/0","type":"type","doc":"Type type for a block of expression in Erlang Abstract Form"},{"ref":"PropCheck.Instrument.html#t:erl_ast_remote_call/0","title":"PropCheck.Instrument.erl_ast_remote_call/0","type":"type","doc":"The type for a remote call in Erlang Abstract Form"},{"ref":"PropCheck.OutputAgent.html","title":"PropCheck.OutputAgent","type":"module","doc":"An agent to gather unique PropCheck-internal output from tests."},{"ref":"PropCheck.OutputAgent.html#child_spec/1","title":"PropCheck.OutputAgent.child_spec/1","type":"function","doc":"Returns a specification to start this module under a supervisor. See Supervisor."},{"ref":"PropCheck.OutputAgent.html#close/1","title":"PropCheck.OutputAgent.close/1","type":"function","doc":"Stop the agent and retrieve the output."},{"ref":"PropCheck.OutputAgent.html#put/2","title":"PropCheck.OutputAgent.put/2","type":"function","doc":"Put new output to the agent."},{"ref":"PropCheck.OutputAgent.html#start_link/0","title":"PropCheck.OutputAgent.start_link/0","type":"function","doc":"Start the agent."},{"ref":"PropCheck.Properties.html","title":"PropCheck.Properties","type":"module","doc":"This module defines the property/4 and property/1 macros. It is automatically available by use PropCheck."},{"ref":"PropCheck.Properties.html#property/1","title":"PropCheck.Properties.property/1","type":"macro","doc":"Defines a not yet implemented property. This convenient macro provides a property which will always flunk. It resembles the test/1 macro from ExUnit. Similarly to ExUnit, it also tags the test with :not_implemented, allowing to filter it when running mix test. Example property &quot;This property will be implemented in the future&quot;"},{"ref":"PropCheck.Properties.html#property/4","title":"PropCheck.Properties.property/4","type":"macro","doc":"Defines a property as part of an ExUnit test. The property macro takes at minimum a name and a do-block containing the code of the property to be tested. The property code is encapsulated as an ExUnit test case of category property, which is released as part of Elixir 1.3 and allows a nice mix of regular unit test and property based testing. This is the reason for the third parameter taking an environment of variables defined in a test setup function. In ExUnit, this is referred to as a test&#39;s &quot;context&quot;. The second parameter sets options for Proper (see PropCheck ). The default is :quiet such that execution during ExUnit runs are silent, as normal unit tests are. You can change it e.g. to :verbose or setting the maximum size of the test data generated or what ever may be helpful. For seeing the result of wrapper functions PropCheck.aggregate/2 etc., the verbose mode is required. Counter Examples If a property fails, the counter example is in a file. The next time this property is checked again, only the counter example is used to ensure that the property now behaves correctly. Additionally, a property with an existing counter example is embellished with the tag failing_prop. You can skip all other tests and property by running mix test --only failing_prop. In this case only the properties with counter example are run. Another option is to use the --stale option of ExUnit to reduce the amount of tests and properties while fixing the code tested by a property. After a property was ran successfully against a previous counter example, PropCheck will run the property again to check if other counter examples can be found. Disable Storing Counter Examples Storing counter examples can be disabled using the :store_counter_example tag. This can be done in three different scopes: module-wide scope, describe-wide scope or for a single property. NOTE that this facility is meant for properties which cannot run with a value generated in a previous test run. This should usually not be the case, and :store_counter_example should only be used after careful consideration. Disable for all properties in a module: defmodule Test do # ... @moduletag store_counter_example: false #... end Disable for all properties in a describe block: defmodule Test do # ... describe &quot;describe block&quot; do @describetag store_counter_example: false # ... end end Disable for a single property: @tag store_counter_example: false property &quot;a property&quot; do # ... end"},{"ref":"PropCheck.StateM.html","title":"PropCheck.StateM","type":"behaviour","doc":"This module defines the :proper_statem behaviour, useful for testing stateful reactive systems whose internal state and side-effects are specified via an abstract state machine. Given a callback module implementing the :proper_statem behaviour (i.e. defining an abstract state machine of the system under test), PropEr can generate random symbolic sequences of calls to that system. As a next step, generated symbolic calls are actually performed, while monitoring the system&#39;s responses to ensure it behaves as expected. Upon failure, the shrinking mechanism attempts to find a minimal sequence of calls provoking the same error. The role of commands Test cases generated for testing a stateful system are lists of symbolic API calls to that system. Symbolic representation has several benefits, which are listed here in increasing order of importance: Generated test cases are easier to read and understand. Failing test cases are easier to shrink. The generation phase is side-effect free and this results in repeatable test cases, which is essential for correct shrinking. Since the actual results of symbolic calls are not known at generation time, we use symbolic variables of type symbolic_var/0 to refer to them. A command of type command/0 is a symbolic term, used to bind a symbolic variable to the result of a symbolic call. For example: [{:set, {:var, 1}, {:call, :erlang, :put, [:a, 42]}}, {:set, {:var, 2}, {:call, :erlang, :erase, [:a]}}, {:set, {:var, 3}, {:call, :erlang, :put, [:b, {:var, 2}]}}] is a command sequence that could be used to test the process dictionary. In this example, the first call stores the pair {:a, 42} in the process dictionary, while the second one deletes it. Then, a new pair {:b, {:var, 2}} is stored. {:var, 2} is a symbolic variable bound to the result of :erlang.erase/1. This result is not known at generation time, since none of these operations is performed at that time. After evaluating the command sequence at runtime, the process dictionary will eventually contain the pair {:b, 42}. The abstract model-state In order to be able to test impure code, we need a way to track its internal state (at least the useful part of it). To this end, we use an abstract state machine representing the possible configurations of the system under test. When referring to the model state, we mean the state of the abstract state machine. The model state can be either symbolic or dynamic: During command generation, we use symbolic variables to bind the results of symbolic calls. Therefore, the model state might (and usually does) contain symbolic variables and/or symbolic calls, which are necessary to operate on symbolic variables. Thus, we refer to it as symbolic state. For example, assuming that the internal state of the process dictionary is modeled as a proplist, the model state after generating the previous command sequence will be [b: {:var, 2}}]. During runtime, symbolic calls are evaluated and symbolic variables are replaced by their corresponding real values. Now we refer to the state as dynamic state. After running the previous command sequence, the model state will be [b: 42]. The callback functions The following functions must be exported from the callback module implementing the abstract state machine: initial_state/0 command/1 precondition/2 postcondition/3 next_state/3 The property used Each test consists of two phases: As a first step, PropEr generates random symbolic command sequences deriving information from the callback module implementing the abstract state machine. This is the role of commands/1 generator. As a second step, command sequences are executed so as to check that the system behaves as expected. This is the role of run_commands/2, a function that evaluates a symbolic command sequence according to an abstract state machine specification. These two phases are encapsulated in the following property, which can be used for testing the process dictionary: def prop_pdict() do forall cmds &lt;- commands(__MODULE__) do {_history, _state, result} = run_commands(__MODULE__, cmds) cleanup() result == ok end end When testing impure code, it is very important to keep each test self-contained. For this reason, almost every property for testing stateful systems contains some clean-up code. Such code is necessary to put the system in a known state, so that the next test can be executed independently from previous ones. Parallel testing After ensuring that a system&#39;s behaviour can be described via an abstract state machine when commands are executed sequentially, it is possible to move to parallel testing. The same state machine can be used to generate command sequences that will be executed in parallel to test for race conditions. A parallel test case (parallel_testcase/0 ) consists of a sequential and a parallel component. The sequential component is a command sequence that is run first to put the system in a random state. The parallel component is a list containing 2 command sequences to be executed in parallel, each of them in a separate newly-spawned process. Generating parallel test cases involves the following actions. Initially, we generate a command sequence deriving information from the abstract state machine specification, as in the case of sequential statem testing. Then, we parallelize a random suffix (up to 12 commands) of the initial sequence by splitting it into 2 subsequences that will be executed concurrently. Limitations arise from the fact that each subsequence should be a valid command sequence (i.e. all commands should satisfy preconditions and use only symbolic variables bound to the results of preceding calls in the same sequence). Furthermore, we apply an additional check: we have to ensure that preconditions are satisfied in all possible interleavings of the concurrent tasks. Otherwise, an exception might be raised during parallel execution and lead to unexpected (and unwanted) test failure. In case these constraints cannot be satisfied for a specific test case, the test case will be executed sequentially. Then an f is printed on screen to inform the user. This usually means that preconditions need to become less strict for parallel testing to work. After running a parallel test case, PropEr uses the state machine specification to check if the results observed could have been produced by a possible serialization of the parallel component. If no such serialization is possible, then an atomicity violation has been detected. In this case, the shrinking mechanism attempts to produce a counterexample that is minimal in terms of concurrent operations. Properties for parallel testing are very similar to those used for sequential testing. def prop_parallel_testing() do forall testcase &lt;- parallel_commands(__MODULE__) do {_sequential, _parallel, result} = run_parallel_commands(__MODULE__, testcase), cleanup(), result == :ok end end Please note that the actual interleaving of commands of the parallel component depends on the Erlang scheduler, which is too deterministic. For PropEr to be able to detect race conditions, the code of the system under test should be instrumented with erlang:yield/0 calls to the scheduler. Acknowledgments Very much of the documentation is immediately taken from the proper API documentation."},{"ref":"PropCheck.StateM.html#c:command/1","title":"PropCheck.StateM.command/1","type":"callback","doc":"Generates a symbolic call to be included in the command sequence, given the current state s of the abstract state machine. However, before the call is actually included, a precondition is checked. This function will be repeatedly called to produce the next call to be included in the test case."},{"ref":"PropCheck.StateM.html#command_names/1","title":"PropCheck.StateM.command_names/1","type":"function","doc":"Extracts the names of the commands from a given command sequence, in the form of MFAs. It is useful in combination with functions such as PropCheck.aggregate/2 in order to collect statistics about command execution."},{"ref":"PropCheck.StateM.html#commands/1","title":"PropCheck.StateM.commands/1","type":"function","doc":"A special PropEr type which generates random command sequences, according to an abstract state machine specification. The function takes as input the name of a callback module, which contains the state machine specification. The initial state is computed by mod:initial_state/0."},{"ref":"PropCheck.StateM.html#commands/2","title":"PropCheck.StateM.commands/2","type":"function","doc":"Similar to commands/1, but generated command sequences always start at a given state. In this case, the first command is always {:init, initial_state} and is used to correctly initialize the state every time the command sequence is run (i.e. during normal execution, while shrinking and when checking a counterexample). In this case, mod:initial_state/0 is never called."},{"ref":"PropCheck.StateM.html#c:initial_state/0","title":"PropCheck.StateM.initial_state/0","type":"callback","doc":"Specifies the symbolic initial state of the state machine. This state will be evaluated at command execution time to produce the actual initial state. The function is not only called at command generation time, but also in order to initialize the state every time the command sequence is run (i.e. during normal execution, while shrinking and when checking a counterexample). For this reason, it should be deterministic and self-contained."},{"ref":"PropCheck.StateM.html#more_commands/2","title":"PropCheck.StateM.more_commands/2","type":"function","doc":"Increases the expected length of command sequences generated from cmd_type by a factor n. CAVEAT This function does not work properly. My current guess is that this is a limitation of how PropEr works with sizing an din particular resizing. The commands list generator (cmd_type) is not a simple list which can be sized easily, but a complex construct where the rather simple approach of resizing does not work as expected."},{"ref":"PropCheck.StateM.html#c:next_state/3","title":"PropCheck.StateM.next_state/3","type":"callback","doc":"Specifies the next state of the abstract state machine, given the current state s, the symbolic call chosen and its result res. This function is called both at command generation and command execution time in order to update the model state, therefore the state s and the result Res can be either symbolic or dynamic."},{"ref":"PropCheck.StateM.html#parallel_commands/1","title":"PropCheck.StateM.parallel_commands/1","type":"function","doc":"A special PropEr type which generates parallel test cases, according to an abstract state machine specification. The function takes as input the name of a callback module, which contains the state machine specification. The initial state is computed by mod:initial_state/0."},{"ref":"PropCheck.StateM.html#parallel_commands/2","title":"PropCheck.StateM.parallel_commands/2","type":"function","doc":"Similar to parallel_commands/1, but generated command sequences always start at a given state."},{"ref":"PropCheck.StateM.html#c:postcondition/3","title":"PropCheck.StateM.postcondition/3","type":"callback","doc":"Specifies the postcondition that should hold about the result res of performing call, given the dynamic state s of the abstract state machine prior to command execution. This function is called during runtime, this is why the state is dynamic."},{"ref":"PropCheck.StateM.html#c:precondition/2","title":"PropCheck.StateM.precondition/2","type":"callback","doc":"Specifies the precondition that should hold so that call can be included in the command sequence, given the current state s of the abstract state machine. In case precondition doesn&#39;t hold, a new call is chosen using the command/1 generator. If preconditions are very strict, it will take a lot of tries for PropEr to randomly choose a valid command. Testing will be stopped in case the constraint_tries limit is reached (see the Options section in the {@link proper} module documentation). Preconditions are also important for correct shrinking of failing test cases. When shrinking command sequences, we try to eliminate commands that do not contribute to failure, ensuring that all preconditions still hold. Validating preconditions is necessary because during shrinking we usually attempt to perform a call with the system being in a state different from the state it was when initially running the test."},{"ref":"PropCheck.StateM.html#print_report/3","title":"PropCheck.StateM.print_report/3","type":"function","doc":"Print pretty report of the failed command run. Note that it by default aliases every commands module, so only the last part of the namespace will be used. Accepts options: return_values - whether to print return values after each command run (default true), last_state - whether section with the last state should be present (default true), pre_cmd_state - whether to print state prior to executed command (default false), post_cmd_state - whether to print state post executed command (default true), cmd_args - whether to print command arguments as literals (default true), inspect_opts - options passed to inspect/2 alias - set a list of modules and aliases, to be used to pretty print commands. Accepts a module/0, a tuple of type {module(), alias :: module()} or a list of them (default is to alias every command&#39;s module - if you want to disable default aliasing set it to [])."},{"ref":"PropCheck.StateM.html#run_commands/2","title":"PropCheck.StateM.run_commands/2","type":"function","doc":"Evaluates a given symbolic command sequence cmds according to the state machine specified in mod. The result is a triple of the form {history, dynamic_state, result}, where: history contains the execution history of all commands that were executed without raising an exception. It contains tuples of the form {t:dynamic_state/0, t:term/0}, specifying the state prior to command execution and the actual result of the command. dynamicState contains the state of the abstract state machine at the moment when execution stopped. In case execution has stopped due to a false postcondition, dynamic_state corresponds to the state prior to execution of the last command. result specifies the outcome of command execution. It can be classified in one of the following categories: ok All commands were successfully run and all postconditions were true. initialization error There was an error while evaluating the initial state. postcondition error A postcondition was false or raised an exception. precondition error A precondition was false or raised an exception. exception An exception was raised while running a command."},{"ref":"PropCheck.StateM.html#run_commands/3","title":"PropCheck.StateM.run_commands/3","type":"function","doc":"Similar to run_commands/2, but also accepts an environment, used for symbolic variable evaluation during command execution. The environment consists of {key::atom, value::any} pairs. Keys may be used in symbolic variables (i.e. {:var, key}) within the command sequence cmds. These symbolic variables will be replaced by their corresponding value during command execution."},{"ref":"PropCheck.StateM.html#run_parallel_commands/2","title":"PropCheck.StateM.run_parallel_commands/2","type":"function","doc":"Runs a given parallel test case according to the state machine specified in mod. The result is a triple of the form {sequential_history, parallel_history, result}, where: sequential_history contains the execution history of the sequential component. Parallel_history contains the execution history of each of the concurrent tasks. Result specifies the outcome of the attempt to serialize command execution, based on the results observed. It can be one of the following: ok no_possible_interleaving Since the scheduler of the runtime system is quite predictable, you are advised to add more turbulences in the execution traces by instrumenting the code under test via PropCheck.Instrument. The history shown when the execution fails, does not reveal in which exact interleaving of commands the failure happens but simply shows the executed commands in both processes running concurrently. This is a feature of PropEr, which cannot easily changed."},{"ref":"PropCheck.StateM.html#run_parallel_commands/3","title":"PropCheck.StateM.run_parallel_commands/3","type":"function","doc":"Similar to run_parallel_commands/2, but also accepts an environment used for symbolic variable evaluation, exactly as described in run_commands/3."},{"ref":"PropCheck.StateM.html#state_after/2","title":"PropCheck.StateM.state_after/2","type":"function","doc":"Returns the symbolic state after running a given command sequence, according to the state machine specification found in mod. The commands are not actually executed."},{"ref":"PropCheck.StateM.html#zip/2","title":"PropCheck.StateM.zip/2","type":"function","doc":"Behaves exactly like Enum.zip/2. Zipping stops when the shortest list stops. This is useful for zipping a command sequence with its (failing) execution history."},{"ref":"PropCheck.StateM.html#t:command/0","title":"PropCheck.StateM.command/0","type":"type","doc":"A value of type command denotes the execution of a symbolic command and storing its result in a symbolic variable."},{"ref":"PropCheck.StateM.html#t:command_list/0","title":"PropCheck.StateM.command_list/0","type":"type","doc":"A sequence of commands."},{"ref":"PropCheck.StateM.html#t:dynamic_state/0","title":"PropCheck.StateM.dynamic_state/0","type":"type","doc":"A dynamic state can be anything and appears only during phase 2."},{"ref":"PropCheck.StateM.html#t:history/0","title":"PropCheck.StateM.history/0","type":"type","doc":"History of command execution in phase 2. It contains current dynamic state and the result of the call."},{"ref":"PropCheck.StateM.html#t:parallel_history/0","title":"PropCheck.StateM.parallel_history/0","type":"type","doc":"The history of concurrent execution of commands in phase 2."},{"ref":"PropCheck.StateM.html#t:parallel_testcase/0","title":"PropCheck.StateM.parallel_testcase/0","type":"type","doc":"A parallel testcase consists of a sequential and a parallel component. The sequential component is a command sequence that is run first to put the system in a random state. The parallel component is a list containing 2 command sequences to be executed in parallel, each of them in a separate newly-spawned process."},{"ref":"PropCheck.StateM.html#t:result/0","title":"PropCheck.StateM.result/0","type":"type","doc":"The outcome of the command sequence execution."},{"ref":"PropCheck.StateM.html#t:symbolic_call/0","title":"PropCheck.StateM.symbolic_call/0","type":"type","doc":"A symbolic call is the typical mfa-tuple plus the tag :call."},{"ref":"PropCheck.StateM.html#t:symbolic_state/0","title":"PropCheck.StateM.symbolic_state/0","type":"type","doc":"A symbolic state can be anything and appears only during phase 1."},{"ref":"PropCheck.StateM.html#t:symbolic_var/0","title":"PropCheck.StateM.symbolic_var/0","type":"type","doc":"Each result of a symbolic call is stored in a symbolic variable. Their values are opaque and can only used as whole."},{"ref":"PropCheck.StateM.DSL.html","title":"PropCheck.StateM.DSL","type":"behaviour","doc":"DEPRECATED : This module is deprecated, please use PropCheck.Statem.ModelDSL instead. This module provides a shallow DSL (domain specific language) in Elixir for property based testing of stateful systems. The basic approach Property based testing of stateful systems is different from ordinary property based testing. Instead of testing operations and their effects on the data structure directly, we construct a model of the system and generate a sequence of commands operating on both, the model and the system. Then we check that after each command step, the system has evolved accordingly to the model. This is the same idea which is used in model checking and is sometimes called a bisimulation. After defining a model, we have two phases during executing the property. In phase 1, the generators create a list of (symbolic) commands including their parameters to be run against the system under test (SUT). A state machine guides the generation of commands. In phase 2, the commands are executed and the state machine checks that the SUT is in the same state as the state machine. If an invalid state is detected, then the command sequence is shrunk towards a shorter sequence serving then as counterexamples. This approach works exactly the same as with PropCheck.StateM and PropCheck.FSM. The main difference is the API, grouping pre- and postconditions, state transitions, and argument generators around the commands of the SUT. This leads towards more logical locality compared to the former implementations. QuickCheck EQC has a similar approach for structuring their modern state machines. The DSL A state machine acting as a model of the SUT can be defined by focusing on states or on transitions. We focus here on the transitions. A transition is a command calling the SUT. Therefore the main phrase of the DSL is the defcommand macro. defcommand :find do # define the rules for executing the find command here end Inside the command macro, we define all the rules which the command must obey. As an example, we discuss here as an example the slightly simplified command :find from test/cache_dsl_test.exs. The SUT is a cache implementation based on an ETS and the model is is based on a list of (key/value)-pairs. This example is derived from Fred Hebert&#39;s PropEr Testing, Chapter 9 The find-command is a call to the find/1 API function. Its arguments are generated by key(), which boils down to numeric values. The arguments for the command are defined by the function args(state) returning a list of generators. In our example, the arguments do not depend on the model state. Next, we need to define the execution of the command by defining function impl/n. This function takes as many arguments as args/1 has elements in the argument list. The impl-function allows to apply conversion of parameters and return values to ease the testing. A typical example is the conversion of an {:ok, value} tuple to only value which can simplify working with value. defcommand :find do def impl(key), do: Cache.find(key) def args(_state), do: [key()] end After defining how a command is executed, we need to define in which state this is allowed. For this, we define function pre/2, taking the model state and the generated list of arguments to check whether this call is allowed in the current model state. In this particular example, find is always allowed, hence we return true without any further checking. This is also the default implementation and the reason why the precondition is missing in the test file. defcommand :find do def impl(key), do: Cache.find(key) def args(_state), do: [key()] def pre(_state, [_key]), do: true end If the precondition is satisfied, the call can happen. After the call, the SUT can be in a different state and the model state must be updated according to the mapping of the SUT to the model. The function next/3 takes the state before the call, the list of arguments and the symbolic or dynamic result (depending on phase 1 or 2, respectively). next/3 returns the new model state. Since searching for a key in the cache does not modify the system nor the model state, nothing has to be done. This is again the default implementation and thus left out in the test file. defcommand :find do def impl(key), do: Cache.find(key) def args(_state), do: [key()] def pre(_state, [_key]), do: true def next(old_state, _args, call_result), do: old_state end The missing part of the command definition is the post condition, checking that after calling the system in phase 2 the system is in the expected state compared the model. This check is implemented in function post/3, which again has a trivial default implementation for post conditions that always returns true. In this example, we check if the call_result is {:error, :not_found}, then we also do not find the key in our model list entries. The other case is that if we a return value of {:ok, val}, then we also find the value via the key in our list of entries. defcommand :find do def impl(key), do: Cache.find(key) def args(_state), do: [key()] def pre(_state, [_key]), do: true def next(old_state, _args, _call_result), do: old_state def post(entries, [key], call_result) do case List.keyfind(entries, key, 0, false) do false -&gt; call_result == {:error, :not_found} {^key, val} -&gt; call_result == {:ok, val} end end end This completes the DSL for command definitions. Additional model elements In addition to commands, we need to define the model itself. This is the ingenious part of stateful property based testing! The initial state of the model must be implemented as the function initial_state/0. From this function, all model evolutions start. In our simplified cache example the initial model is an empty list: def initial_state(), do: [] The commands are generated with the same frequency by default. Often, this is not appropriate, e.g. in the cache example we expect many more find than cache commands. Therefore, commands can have a weight, which is technically used inside a PropCheck.BasicTypes.frequency/1 generator. The weights are defined in callback function weight/1, taking the current model state and returning a map of command and frequency pairs to be generated. In our cache example we want the find command to appear three times more often than other commands: def weight(_state), do: %{find: 3, cache: 1, flush: 1} The property to test The property to test the stateful system is more or less the same for all systems. We generate all commands via generator commands/1, which takes a module with callbacks as parameter. Inside the test, we first start the SUT, execute the commands with run_commands/1, stopping the SUT and evaluating the result of the executions as a boolean expression. This boolean expression can be adorned with further functions and macros to analyze the generated commands (via PropCheck.aggregate/2) or to inspect the history if a failure occurs (via PropCheck.when_fail/2). In the test cases, you find more examples of such adornments. property &quot;run the sequential cache&quot;, [:verbose] do forall cmds &lt;- commands(__MODULE__) do Cache.start_link(@cache_size) execution = run_commands(cmds) Cache.stop() (execution.result == :ok) end end Increasing the Number of Commands in a Sequence Sometimes issues can hide when the command sequences are short. In order to tease out these hidden bugs we can increase the number of commands generated by using the max_size option in our property. property &quot;run the sequential cache&quot;, [max_size: 250] do forall cmds &lt;- commands(__MODULE__) do Cache.start_link(@cache_size) execution = run_commands(cmds) Cache.stop() (execution.result == :ok) end"},{"ref":"PropCheck.StateM.DSL.html#command_names/1","title":"PropCheck.StateM.DSL.command_names/1","type":"function","doc":"Takes a list of generated commands and returns a list of mfa-tuples. This can be used for aggregation of commands."},{"ref":"PropCheck.StateM.DSL.html#commands/1","title":"PropCheck.StateM.DSL.commands/1","type":"function","doc":"Generates the command list for the given module"},{"ref":"PropCheck.StateM.DSL.html#defcommand/2","title":"PropCheck.StateM.DSL.defcommand/2","type":"macro","doc":"DEPRECATED : This module is deprecated, please use PropCheck.Statem.ModelDSL instead. Defines a new command of the model. Inside the command, local functions define how the command is executed (impl(...)). This is required. how the arguments in the current model state are generated (args(state). The default is the empty list of arguments. if the command is allowed in the current model state (pre(state, arg_list) :: boolean) This is true per default. what the next state of the model is after the call (next(old_state, arg_list, result) :: new_state). The default implementation does not change the model state, sufficient for queries. if the system under test is in the correct state after the call (post(old_state, arg_list, result) :: boolean). This is true in the default implementation. These local functions inside the macro are effectively callbacks to guide and evolve the model state."},{"ref":"PropCheck.StateM.DSL.html#c:initial_state/0","title":"PropCheck.StateM.DSL.initial_state/0","type":"callback","doc":"The initial state of the state machine is computed by this callback."},{"ref":"PropCheck.StateM.DSL.html#run_commands/1","title":"PropCheck.StateM.DSL.run_commands/1","type":"function","doc":"Runs the list of generated commands according to the model. Returns the result, the history and the final state of the model. Due to an internal refactoring and to achieve a common API with the PropCheck.StateM module, we changed the API for run_commands. This implementation infers the callback module from the first generated command. Usually, this will be the case, but we cannot rely on that."},{"ref":"PropCheck.StateM.DSL.html#run_commands/2","title":"PropCheck.StateM.DSL.run_commands/2","type":"function","doc":"Runs the list of generated commands according to the model. Returns the result, the history and the final state of the model."},{"ref":"PropCheck.StateM.DSL.html#c:weight/1","title":"PropCheck.StateM.DSL.weight/1","type":"callback","doc":"DEPRECATED : This module is deprecated, please use PropCheck.Statem.ModelDSL instead. The optional weights for the command generation. It takes the current model state and returns a map of command/weight pairs. Commands, which are not allowed in a specific state, should be omitted, since a frequency of 0 is not allowed. def weight(state), do: %{x: 1, y: 1, a: 2, b: 2}"},{"ref":"PropCheck.StateM.DSL.html#t:command/0","title":"PropCheck.StateM.DSL.command/0","type":"type","doc":"A value of type command denotes the execution of a symbolic command and storing its result in a symbolic variable."},{"ref":"PropCheck.StateM.DSL.html#t:command_name/0","title":"PropCheck.StateM.DSL.command_name/0","type":"type","doc":"The name of a command must be an atom."},{"ref":"PropCheck.StateM.DSL.html#t:dynamic_state/0","title":"PropCheck.StateM.DSL.dynamic_state/0","type":"type","doc":"A dynamic state can be anything and appears only during phase 2."},{"ref":"PropCheck.StateM.DSL.html#t:history_event/0","title":"PropCheck.StateM.DSL.history_event/0","type":"type","doc":"The history of command execution in phase 2 is stored in a history element. It contains the current dynamic state and the call to be made."},{"ref":"PropCheck.StateM.DSL.html#t:result_t/0","title":"PropCheck.StateM.DSL.result_t/0","type":"type","doc":"The result of the command execution. It contains either the state of the failing precondition, the command&#39;s return value of the failing postcondition, the exception values or :ok if everything is fine."},{"ref":"PropCheck.StateM.DSL.html#t:state_call/0","title":"PropCheck.StateM.DSL.state_call/0","type":"type","doc":"The sequence of calls consists of state and symbolic calls."},{"ref":"PropCheck.StateM.DSL.html#t:state_t/0","title":"PropCheck.StateM.DSL.state_t/0","type":"type","doc":"The combination of symbolic and dynamic states are required for functions which are used in both phases 1 and 2."},{"ref":"PropCheck.StateM.DSL.html#t:symbolic_call/0","title":"PropCheck.StateM.DSL.symbolic_call/0","type":"type","doc":"A symbolic call is the typical mfa-tuple plus the indicator :call."},{"ref":"PropCheck.StateM.DSL.html#t:symbolic_state/0","title":"PropCheck.StateM.DSL.symbolic_state/0","type":"type","doc":"A symbolic state can be anything and appears only during phase 1."},{"ref":"PropCheck.StateM.DSL.html#t:symbolic_var/0","title":"PropCheck.StateM.DSL.symbolic_var/0","type":"type","doc":"Each result of a symbolic call is stored in a symbolic variable. Their values are opaque and can only used as whole."},{"ref":"PropCheck.StateM.DSL.html#t:t/0","title":"PropCheck.StateM.DSL.t/0","type":"type","doc":"The combined result of the test. It contains the history of all executed commands, the final state, the final result and the environment, mapping symbolic vars to their actual values. Everything is fine, if result is :ok."},{"ref":"PropCheck.StateM.ModelDSL.html","title":"PropCheck.StateM.ModelDSL","type":"behaviour","doc":"This module provides a shallow DSL (domain specific language) in Elixir for property based testing of stateful systems. It&#39;s built upon PropCheck.StateM and all it&#39;s the characteristics apply here as well. It&#39;s a replacement for PropCheck.StateM.DSL. The basic approach Property based testing of stateful systems is different from ordinary property based testing. Instead of testing operations and their effects on the data structure directly, we construct a model of the system and generate a sequence of commands operating on both, the model and the system. Then we check that after each command step, the system has evolved accordingly to the model. This is the same idea which is used in model checking and is sometimes called a bisimulation. After defining a model, we have two phases during executing the property. In phase 1, the generators create a list of (symbolic) commands including their parameters to be run against the system under test (SUT). A state machine guides the generation of commands. In phase 2, the commands are executed and the state machine checks that the SUT is in the same state as the state machine. If an invalid state is detected, then the command sequence is shrunk towards a shorter sequence serving then as counterexamples. This approach works exactly the same as with PropCheck.StateM and PropCheck.FSM. The main difference is the API, grouping pre- and postconditions, state transitions around the commands of the SUT. This leads towards more logical locality compared to the former implementations. QuickCheck EQC has a similar approach for structuring their modern state machines. The DSL A state machine acting as a model of the SUT can be defined by focusing on states or on transitions. We focus here on the transitions. A transition is a command calling the SUT. Therefore the main phrase of the DSL is the defcommand macro. defcommand :find do # define the rules for executing the find command here end Inside the defcommand macro, we define all the rules which the command must obey. As an example, we discuss here as an example the slightly simplified command :find from test/cache_dsl_test.exs. The SUT is a cache implementation based on an ETS and the model is is based on a list of (key/value)-pairs. This example is derived from Fred Hebert&#39;s PropEr Testing, Chapter 9 The find-command is a call to the find/1 API function. Its arguments are generated in command_gen/1 (described later) callback, which for this command is using just one argument, a key() generator. Next, we need to define the execution of the command by defining function impl/n. The impl-function allows to apply conversion of parameters and return values to ease the testing. A typical example is the conversion of an {:ok, value} tuple to only value which can simplify working with value. defcommand :find do def impl(key), do: Cache.find(key) end After defining how a command is executed, we need to define in which state this is allowed. For this, we define function pre/2, taking the model state and the generated list of arguments to check whether this call is allowed in the current model state. In this particular example, find is always allowed, hence we return true without any further checking. This is also the default implementation and the reason why the precondition is missing in the test file. defcommand :find do def impl(key), do: Cache.find(key) def pre(_state, [_key]), do: true end If the precondition is satisfied, the call can happen. After the call, the SUT can be in a different state and the model state must be updated according to the mapping of the SUT to the model. The function next/3 takes the state before the call, the list of arguments and the symbolic or dynamic result (depending on phase 1 or 2, respectively). next/3 returns the new model state. Since searching for a key in the cache does not modify the system nor the model state, nothing has to be done. This is again the default implementation and thus left out in the test file. defcommand :find do def impl(key), do: Cache.find(key) def pre(_state, [_key]), do: true def next(old_state, _args, call_result), do: old_state end The missing part of the command definition is the post condition, checking that after calling the system in phase 2, the system is in the expected state compared the model. This check is implemented in function post/3, which again has a trivial default implementation for post conditions that always returns true. In this example, we check if we can find the key in our list of entries and if we do, we check if call_result resulted in {:ok, val}. Or if we don&#39;t found it, we check if the SUT also cannot find it by comparing if call_result returned {:error, :not_found}. defcommand :find do def impl(key), do: Cache.find(key) def pre(_state, [_key]), do: true def next(old_state, _args, _call_result), do: old_state def post(entries, [key], call_result) do case List.keyfind(entries, key, 0, false) do false -&gt; call_result == {:error, :not_found} {^key, val} -&gt; call_result == {:ok, val} end end end This completes the DSL for command definitions. Additional model elements In addition to commands, we need to define the model itself. This is the ingenious part of stateful property based testing! The initial state of the model must be implemented as the function initial_state/0. It doesn&#39;t accept any arguments, because this function has to be deterministic. From this function, all model evolutions start. In our simplified cache example the initial model is an empty list: def initial_state(), do: [] The sequence of commands to be run is generated repeatedly by command_gen/1 callback. The generator has to return a tuple of with a command name and a list of it&#39;s arguments (a list of generators). This callback expects the current state as an argument, which often is used to determine the next one from a set of appropriate commands (e.g. there might not be much sense in calling the delete_user command, if there are no users in the system yet). Usually a PropCheck.BasicTypes.oneof/1 or PropCheck.BasicTypes.frequency/1 generators are used to pick one of possible commands. In our cache example we want the find command to appear three times more often than other commands: def command_gen(_state) do frequency([ {3, {:find, [key()]}}, {1, {:cache, [key(), val()]}}, {1, {:flush, []}} ]) The property to test The property to test the stateful system is more or less the same for all systems. We generate all commands via generator commands/1, which takes a module with callbacks as parameter. Inside the test, we first start the SUT, execute the commands with run_commands/1, stopping the SUT and evaluating the result of the executions as a boolean expression. This boolean expression can be adorned with further functions and macros to analyze the generated commands (via PropCheck.aggregate/2) or to inspect the history if a failure occurs (via PropCheck.when_fail/2). In the test cases, you find more examples of such adornments. property &quot;run the sequential cache&quot;, [:verbose] do forall cmds &lt;- commands(__MODULE__) do Cache.start_link(@cache_size) {_history, _state, result} = run_commands(cmds) Cache.stop() (result == :ok) end end Increasing the Number of Commands in a Sequence Sometimes issues can hide when the command sequences are short. In order to tease out these hidden bugs we can increase the number of commands generated by using the max_size option in our property. property &quot;run the sequential cache&quot;, [max_size: 250] do forall cmds &lt;- commands(__MODULE__) do Cache.start_link(@cache_size) {_history, _state, result} = run_commands(cmds) Cache.stop() (result == :ok) end"},{"ref":"PropCheck.StateM.ModelDSL.html#c:command_gen/1","title":"PropCheck.StateM.ModelDSL.command_gen/1","type":"callback","doc":"Generates a symbolic call to be included in the command sequence, given the current state s of the abstract state machine. Must return a type that generates tuples {command_name :: atom, args :: [PropCheck.BasicTypes.type]}. However, before the call is actually included, a precondition is checked. This function will be repeatedly called to produce the next call to be included in the test case."},{"ref":"PropCheck.StateM.ModelDSL.html#command_names/1","title":"PropCheck.StateM.ModelDSL.command_names/1","type":"function","doc":"Extracts the names of the commands from a given command sequence, in the form of MFAs. It is useful in combination with functions such as PropCheck.aggregate/2 in order to collect statistics about command execution."},{"ref":"PropCheck.StateM.ModelDSL.html#commands/1","title":"PropCheck.StateM.ModelDSL.commands/1","type":"function","doc":"A special PropEr type which generates random command sequences, according to an abstract state machine specification. The function takes as input the name of a callback module, which contains the state machine specification. The initial state is computed by mod:initial_state/0."},{"ref":"PropCheck.StateM.ModelDSL.html#commands/2","title":"PropCheck.StateM.ModelDSL.commands/2","type":"function","doc":"Similar to commands/1, but generated command sequences always start at a given state. In this case, the first command is always {:init, initial_state} and is used to correctly initialize the state every time the command sequence is run (i.e. during normal execution, while shrinking and when checking a counterexample). In this case, mod:initial_state/0 is never called."},{"ref":"PropCheck.StateM.ModelDSL.html#def_commands/0","title":"PropCheck.StateM.ModelDSL.def_commands/0","type":"function","doc":""},{"ref":"PropCheck.StateM.ModelDSL.html#def_next_states/1","title":"PropCheck.StateM.ModelDSL.def_next_states/1","type":"function","doc":""},{"ref":"PropCheck.StateM.ModelDSL.html#def_postconds/1","title":"PropCheck.StateM.ModelDSL.def_postconds/1","type":"function","doc":""},{"ref":"PropCheck.StateM.ModelDSL.html#def_preconds/1","title":"PropCheck.StateM.ModelDSL.def_preconds/1","type":"function","doc":""},{"ref":"PropCheck.StateM.ModelDSL.html#defcommand/2","title":"PropCheck.StateM.ModelDSL.defcommand/2","type":"macro","doc":"Defines a new command of the model. Inside the command, local functions define how the command is executed: impl(...) - this is required, if the command is allowed in the current model state: pre(state, arg_list) :: boolean - this is true per default, what the next state of the model is after the call: next(old_state, arg_list, result) :: new_state - the default implementation does not change the model state, sufficient for queries, if the system under test is in the correct state after the call: post(old_state, arg_list, result) :: boolean - this is true in the default implementation. These local functions inside the macro are effectively callbacks to guide and evolve the model state."},{"ref":"PropCheck.StateM.ModelDSL.html#c:initial_state/0","title":"PropCheck.StateM.ModelDSL.initial_state/0","type":"callback","doc":"Specifies the symbolic initial state of the state machine. This state will be evaluated at command execution time to produce the actual initial state. The function is not only called at command generation time, but also in order to initialize the state every time the command sequence is run (i.e. during normal execution, while shrinking and when checking a counterexample). For this reason, it should be deterministic and self-contained."},{"ref":"PropCheck.StateM.ModelDSL.html#more_commands/2","title":"PropCheck.StateM.ModelDSL.more_commands/2","type":"function","doc":"Increases the expected length of command sequences generated from cmd_type by a factor n. CAVEAT This function does not work properly. My current guess is that this is a limitation of how PropEr works with sizing an din particular resizing. The commands list generator (cmd_type) is not a simple list which can be sized easily, but a complex construct where the rather simple approach of resizing does not work as expected."},{"ref":"PropCheck.StateM.ModelDSL.html#parallel_commands/1","title":"PropCheck.StateM.ModelDSL.parallel_commands/1","type":"function","doc":"A special PropEr type which generates parallel test cases, according to an abstract state machine specification. The function takes as input the name of a callback module, which contains the state machine specification. The initial state is computed by mod:initial_state/0."},{"ref":"PropCheck.StateM.ModelDSL.html#parallel_commands/2","title":"PropCheck.StateM.ModelDSL.parallel_commands/2","type":"function","doc":"Similar to parallel_commands/1, but generated command sequences always start at a given state."},{"ref":"PropCheck.StateM.ModelDSL.html#print_report/3","title":"PropCheck.StateM.ModelDSL.print_report/3","type":"function","doc":"Print pretty report of the failed command run. Note that it by default aliases every commands module, so only the last part of the namespace will be used. Accepts options: return_values - whether to print return values after each command run (default true), last_state - whether section with the last state should be present (default true), pre_cmd_state - whether to print state prior to executed command (default false), post_cmd_state - whether to print state post executed command (default true), cmd_args - whether to print command arguments as literals (default true), inspect_opts - options passed to inspect/2 alias - set a list of modules and aliases, to be used to pretty print commands. Accepts a module/0, a tuple of type {module(), alias :: module()} or a list of them (default is to alias every command&#39;s module - if you want to disable default aliasing set it to [])."},{"ref":"PropCheck.StateM.ModelDSL.html#run_commands/2","title":"PropCheck.StateM.ModelDSL.run_commands/2","type":"function","doc":"Evaluates a given symbolic command sequence cmds according to the state machine specified in mod. The result is a triple of the form {history, dynamic_state, result}, where: history contains the execution history of all commands that were executed without raising an exception. It contains tuples of the form {t:dynamic_state, t:term}, specifying the state prior to command execution and the actual result of the command. dynamicState contains the state of the abstract state machine at the moment when execution stopped. In case execution has stopped due to a false postcondition, dynamic_state corresponds to the state prior to execution of the last command. result specifies the outcome of command execution. It can be classified in one of the following categories: ok All commands were successfully run and all postconditions were true. initialization error There was an error while evaluating the initial state. postcondition error A postcondition was false or raised an exception. precondition error A precondition was false or raised an exception. exception An exception was raised while running a command."},{"ref":"PropCheck.StateM.ModelDSL.html#run_commands/3","title":"PropCheck.StateM.ModelDSL.run_commands/3","type":"function","doc":"Similar to run_commands/2, but also accepts an environment, used for symbolic variable evaluation during command execution. The environment consists of {key::atom, value::any} pairs. Keys may be used in symbolic variables (i.e. {:var, key}) within the command sequence cmds. These symbolic variables will be replaced by their corresponding value during command execution."},{"ref":"PropCheck.StateM.ModelDSL.html#run_parallel_commands/2","title":"PropCheck.StateM.ModelDSL.run_parallel_commands/2","type":"function","doc":"Runs a given parallel test case according to the state machine specified in mod. The result is a triple of the form {sequential_history, parallel_history, result}, where: sequential_history contains the execution history of the sequential component. Parallel_history contains the execution history of each of the concurrent tasks. Result specifies the outcome of the attempt to serialize command execution, based on the results observed. It can be one of the following: ok no_possible_interleaving Since the scheduler of the runtime system is quite predictable, you are advised to add more turbulences in the execution traces by instrumenting the code under test via PropCheck.Instrument. The history shown when the execution fails, does not reveal in which exact interleaving of commands the failure happens but simply shows the executed commands in both processes running concurrently. This is a feature of PropEr, which cannot easily changed."},{"ref":"PropCheck.StateM.ModelDSL.html#run_parallel_commands/3","title":"PropCheck.StateM.ModelDSL.run_parallel_commands/3","type":"function","doc":"Similar to run_parallel_commands/2, but also accepts an environment used for symbolic variable evaluation, exactly as described in run_commands/3."},{"ref":"PropCheck.StateM.ModelDSL.html#state_after/2","title":"PropCheck.StateM.ModelDSL.state_after/2","type":"function","doc":"Returns the symbolic state after running a given command sequence, according to the state machine specification found in mod. The commands are not actually executed."},{"ref":"PropCheck.StateM.ModelDSL.html#zip/2","title":"PropCheck.StateM.ModelDSL.zip/2","type":"function","doc":"Behaves exactly like Enum.zip/2. Zipping stops when the shortest list stops. This is useful for zipping a command sequence with its (failing) execution history."},{"ref":"PropCheck.StateM.ModelDSL.html#t:command/0","title":"PropCheck.StateM.ModelDSL.command/0","type":"type","doc":"A value of type command denotes the execution of a symbolic command and storing its result in a symbolic variable."},{"ref":"PropCheck.StateM.ModelDSL.html#t:command_list/0","title":"PropCheck.StateM.ModelDSL.command_list/0","type":"type","doc":"A sequence of commands."},{"ref":"PropCheck.StateM.ModelDSL.html#t:dynamic_state/0","title":"PropCheck.StateM.ModelDSL.dynamic_state/0","type":"type","doc":"A dynamic state can be anything and appears only during phase 2."},{"ref":"PropCheck.StateM.ModelDSL.html#t:history/0","title":"PropCheck.StateM.ModelDSL.history/0","type":"type","doc":"History of command execution in phase 2. It contains current dynamic state and the result of the call."},{"ref":"PropCheck.StateM.ModelDSL.html#t:parallel_history/0","title":"PropCheck.StateM.ModelDSL.parallel_history/0","type":"type","doc":"The history of concurrent execution of commands in phase 2."},{"ref":"PropCheck.StateM.ModelDSL.html#t:parallel_testcase/0","title":"PropCheck.StateM.ModelDSL.parallel_testcase/0","type":"type","doc":"A parallel testcase consists of a sequential and a parallel component. The sequential component is a command sequence that is run first to put the system in a random state. The parallel component is a list containing 2 command sequences to be executed in parallel, each of them in a separate newly-spawned process."},{"ref":"PropCheck.StateM.ModelDSL.html#t:result/0","title":"PropCheck.StateM.ModelDSL.result/0","type":"type","doc":"The outcome of the command sequence execution."},{"ref":"PropCheck.StateM.ModelDSL.html#t:symbolic_call/0","title":"PropCheck.StateM.ModelDSL.symbolic_call/0","type":"type","doc":"A symbolic call is the typical mfa-tuple plus the tag :call."},{"ref":"PropCheck.StateM.ModelDSL.html#t:symbolic_state/0","title":"PropCheck.StateM.ModelDSL.symbolic_state/0","type":"type","doc":"A symbolic state can be anything and appears only during phase 1."},{"ref":"PropCheck.StateM.ModelDSL.html#t:symbolic_var/0","title":"PropCheck.StateM.ModelDSL.symbolic_var/0","type":"type","doc":"Each result of a symbolic call is stored in a symbolic variable. Their values are opaque and can only used as whole."},{"ref":"PropCheck.TargetedPBT.html","title":"PropCheck.TargetedPBT","type":"module","doc":"This module defines the top-level behaviour for targeted property-based testing (TPBT). Using TPBT the input generation is no longer random, but guided by a search strategy to increase the probability of finding failing input. For this to work the user has to specify a search strategy and also needs to extract utility-values from the system under test that the search strategy then tries to maximize. To use TPBT the test specification macros forall_targeted, exists, and not_exists are used. The typical structure for a targeted property looks as follows: property prop_target do # Try to check that exists input &lt;- params do # some input exists that fulfills the property. uv = sut.run(input) # Do so by running SUT with input maximize(uv) # and maximize its utility value uv &lt; threshold # up to some threshold, the property condition. end end Some thoughts on strategies how to use targeted properties Targeted PBT is a rather new technology and really fascinating. But when should you to use it and when are the classical technologies more suitable? Here are some thoughts on that topic based on the limited experience we currently have. If you want to test interesting parts of your algorithms or data structures, you would typically invest into more elaborated generators. If you are working on binary trees and you need more lefty trees, then a new generator is required that somehow generates trees of that particular shape. The big advantage here is that you can use all of the standard functions and macros, in particular for collecting statistics about the quality of the generated data (e.g. by PropCheck.collect/2 and its friends). Another approach would be to use targeted properties as shown in targeted_tree_test.exs: You use a rather simple data generator and define a measuring function on the data to express the &quot;leftiness&quot; of the tree. Equiped with that, a target property searches automatically for data the maximizes (or minimizes) the measuring function (often called utility value or function). Not inventing a clever data generators comes with a price - there is no free lunch... You need some relevant property of your generated data which you can measure, i.e. you need a function from your data to the real numbers. This is not always possible. Searching for an optimum takes more time than simply generating random data: the run-time of the properties increases. Data collecting functions and macros such as PropCheck.collect/2 are currently not available. This it the reason why in targeted_tree_test.exs we use print-outs to show and verify the generated data. Counter examples and shrinking are not available The current implementation in PropEr does not work well together with recursive data generators, which renders the approach unusable for state-based PBT. But, of course, you gain also something. You can use rather straight data generators and let the searching algorithm find the interesting parts with respect to the measuring function. In level_tpbt_test.exs a very different approach is used. Here the basic idea is to verify that a data structure (here: a maze in a computer game) has a proper structure (here: there exists at least one valid path from the entrance to the exit of the maze). The function to minimize is the distance from the end of the path to the exit position. The searching algorithm then optimizes the path length for a minimal distance until the exit is found. For more complex mazes, it is required to adopt the amount of search_steps and the neighbor_hood function to find the exit. They takes over the role numtests and resize to enlarge or refit the generated data for the next search step. You can combine approaches by using a classical generator for e.g. generating a new maze, and then use inside a targeted property to find a path to the maze&#39;s exit. This would be roughly like this: forall maze &lt;- maze_generator() do exists p &lt;- path_generator() do pos = Maze.follow_path(maze, maze.entry_pos, p) uv = distance(pos, maze.exit_pos) minimize(uv) pos == maze.exit_pos end end How the targeted properties relate The targeted macros forall_targeted, exists and not_exists are related to each other. The universal laws of quantors from first-order logic apply here as well (cf. provable entities in first-order logic) and explain why some conditions in the test examples are constructed the way they are. In the following, we use the variable x, the generator x_gen() and a boolean predicate p(). The term &lt;==&gt; means that the expression on both sides are equivalent. forall_targeted x &lt;- x_gen(), do: p(x) &lt;==&gt; not_exists x &lt;- x_gen(), do: not(p(x)) exists x &lt;- x_gen(), do: p(x) &lt;==&gt; forall_targeted x &lt;- x_gen(), do: not(p(x)) |&gt; fails() not_exists x &lt;- x_gen(), do: p(x) &lt;==&gt; forall_targeted x &lt;- x_gen(), do: not(p(x)) Options For targeted properties exists a new option: {:search_steps, non_negative_number} takes an integer defining how many search steps the searching algorithm takes. Its default value is 1_000. The effect of search_steps is similar to num_tests for ordinary properties. num_tests has no effect on the search strategy. This helps when you combine a regular property with search strategy, e.g. generating a configuration parameter and search for specific properties to hold depending on that parameter. Some of the documentation is taken directly from PropEr."},{"ref":"PropCheck.TargetedPBT.html#exists/2","title":"PropCheck.TargetedPBT.exists/2","type":"macro","doc":"The exists macro uses the targeted PBT component of PropEr to try to find one instance of xs that makes the prop return true. If such a xs is found, the property passes. Note that there is no counterexample if no such xs could be found."},{"ref":"PropCheck.TargetedPBT.html#forall_targeted/2","title":"PropCheck.TargetedPBT.forall_targeted/2","type":"macro","doc":"The forall_targeted macros uses the targeted PBT component of PropEr to try that all instances of xs fulfill property prop. In contrast to exists, often the property here is negated."},{"ref":"PropCheck.TargetedPBT.html#maximize/1","title":"PropCheck.TargetedPBT.maximize/1","type":"macro","doc":"This macro tells the search strategy to maximize the value fitness."},{"ref":"PropCheck.TargetedPBT.html#minimize/1","title":"PropCheck.TargetedPBT.minimize/1","type":"macro","doc":"This macro tells the search strategy to minimize the value fitness and is equivalent to maximize(-fitness)."},{"ref":"PropCheck.TargetedPBT.html#not_exists/2","title":"PropCheck.TargetedPBT.not_exists/2","type":"macro","doc":"The not_exists macro uses the targeted PBT component of PropEr to try to find one instance of xs that makes the prop return false. If such a xs is found the property passes. Note that there is no counterexample if no such xs could be found."},{"ref":"PropCheck.TargetedPBT.html#user_matcher/2","title":"PropCheck.TargetedPBT.user_matcher/2","type":"macro","doc":"This overwrites the structural matching of PropEr for the generator with the user provided matcher function. The matcher should be of type :proper_gen_next.matcher()"},{"ref":"PropCheck.TargetedPBT.html#user_nf/2","title":"PropCheck.TargetedPBT.user_nf/2","type":"macro","doc":"This uses the neighborhood function nf instead of PropEr&#39;s constructed neighborhood function for this generator. The neighborhood function nf should be of type fun(any(), {Depth :: number(), Temperature::float()} -&gt; any()"},{"ref":"PropCheck.YieldInstrumenter.html","title":"PropCheck.YieldInstrumenter","type":"module","doc":"Instruments with prepending :erlang.yield/0 for calls typical concurrency bug aware functions."},{"ref":"readme.html","title":"PropCheck - Property based testing for Elixir","type":"extras","doc":"PropCheck - Property based testing for Elixir PropCheck is a testing library, that provides a wrapper around PropEr, an Erlang based property testing framework in the spirit of QuickCheck."},{"ref":"readme.html#installation","title":"PropCheck - Property based testing for Elixir - Installation","type":"extras","doc":"To use PropCheck with your project, add it as a dependency to mix.exs: defp deps do [ {:propcheck, &quot;~&gt; 1.1&quot;, only: [:test, :dev]} ] end"},{"ref":"readme.html#changes","title":"PropCheck - Property based testing for Elixir - Changes","type":"extras","doc":"Relevant changes are documented in the Changelog, on GitHubfollow this link."},{"ref":"readme.html#basic-usage-and-build-configuration","title":"PropCheck - Property based testing for Elixir - Basic Usage and Build Configuration","type":"extras","doc":"PropCheck allows to define properties, which automatically executed via ExUnit when running mix test. You find relevant information here: details about the property macro are found in PropCheck.Properties, details about how to specify the property conditions are documented in PropCheck, the basic data generators are found in PropCheck.BasicTypes, for property testing of state-based systems take a loot at PropCheck.StateM.ModelDSL for the new DSL (since PropCheck 1.1.0-rc1), which is a layer on top of PropCheck.StateM. The new targeted property based testing approach (TBPT) employing an automated search strategy towards more interesting test data is described in PropCheck.TargetedPBT. For PropCheck, there is only one configuration option. All counter examples found are stored in a file, the name of which is configurable in mix.exs as part of the project configuration: def project() do [ # many other options propcheck: [counter_examples: &quot;filename&quot;] ] end Per default, the counter examples file is stored in the build directory (_build), independent from the build environment, in the file propcheck.ctex. The file can be removed using mix propcheck.clean. Note that this task is only available if PropCheck is also available in :dev. Otherwise, MIX_ENV=test mix propcheck.clean can be used. Setting PropCheck into Verbose Mode Properties in PropCheck can be run in quiet or verbose mode. Usually, quiet is the default. To enable verbose mode without requiring to touch the source code, the environment variable PROPCHECK_VERBOSE can be used. If this is set to 1, the forall macro prints verbose output. Detecting Exceptions and Errors in Tests PropCheck can attempt to detect and output exceptions in non-verbose mode. This can be done using the detect_exceptions: true option for property or forall, or using the environment variable PROPCHECK_DETECT_EXCEPTIONS. If this environment variable is set to 1, exception detection is enabled globally."},{"ref":"readme.html#links-to-other-documentation","title":"PropCheck - Property based testing for Elixir - Links to other documentation","type":"extras","doc":"The guides for PropEr are an essential set of documents to make full use of PropCheck PropEr Home Page PropEr Tutorials Elixir versions of most of PropEr&#39;s tutorial material can be found in the test folder on GitHub. Jesper Andersen and Robert Aloi blog about their thoughts and experience on using QuickCheck which are (mostly) directly transferable to PropCheck (with the notable exception of concurrency and the new state machine DSL from QuickCheck with the possibility to add requirement tags): Jesper Andersen&#39;s QuickCheck Advice Jesper Andersen&#39;s Breaking Erlang Maps (4 part series) Roberto Aloi&#39;s Notes on Erlang Quickcheck Rather new introductory resources are Fred Hebert&#39;s PropEr Testing and Fred Hebert&#39;s book Property-Based Testing With PropEr, Erlang and Elixir. This book explains the new approach of targeted property based testing (TPBT) very nicely and is way more approachable then the scientific papers regarding TPBT. Some of the Erlang only examples of the book are ported to Elixir and can be found in the test test folder on GitHub. In contrast to the book, the free website is concerned with Erlang only. The Erlang examples translate easily into Elixir (beside that at least a reading knowledge of Erlang is extremely helpful to survive in the BEAM ecosystem ...). Eventually I will port some of the examples to Elixir and PropCheck and certainly like to accept PRs."},{"ref":"readme.html#what-is-not-available","title":"PropCheck - Property based testing for Elixir - What is not available","type":"extras","doc":"PropCheck does not support PropEr&#39;s capability to derive automatically type generators from type specifications. This is due to some shortcomings in PropEr, where the specification analysis in certain situation attempt to parse the Erlang source code of the file with the type specification. Apparently, this fails if the types are specified in an Elixir source file. Effectively this means, that to to support this feature from Elixir, the type management system in PropEr needs to be rewritten completely. Jesper Andersen argues in his aforementioned blog post eloquently that automatically derived type generators are not needed, even more that carefully crafted type generators for specific testing purposes is an essential part of the QuickCheck philosophy. Therefore, this missing feature is not that bad. For the same reason, automatic @spec-checking is of limited value in PropCheck since type generators for functions specification are also generated automatically. PropCheck has only very limited support for parallel testing since it introduces no new features for concurrency compared to PropEr."},{"ref":"readme.html#contributing","title":"PropCheck - Property based testing for Elixir - Contributing","type":"extras","doc":"Please use the GitHub issue tracker for bug reports and for submitting pull requests Test Policy Before submitting a pull request, please use Credo to ensure code consistency and run mix tests to check PropCheck itself. mix tests is a Mix alias that runs regular tests (via mix test) and some external tests (via another Mix alias mix test_ext) which test PropCheck&#39;s side effects, like storing counterexamples or proper output format, that can&#39;t be easily tested using regular tests. Rebase Policy Ensure that your PR is based on the latest master version by rebasing. If your feature branch is my_feature, then do the following before publishing a pull request: git checkout master git pull --rebase git checkout my_feature git rebase master If your PR implementation takes longer such that another PR is merged before your own PR, then you have to repeat the above sequence. It might be that you have fix some conflicts. But now you cannot push your branch again, because you changed the history of your local branch compared to the one published on GitHub. Therefore, you have force-push your branch. Do this with git push --force-with-lease A simple git push --force is not allowed, --force-with-lease is more friendly and thus allowed. See details in the Git documentation. The rationale behind this policy is that we want a simple almost linear history, where each merged PR create a sequence of merge with no parallel work. This history will not show how many active branches are available during development but the sequence of incorporating changes to master. That is the important part and we want to see which commit sequence lead to the specific feature. Merges destroy this linearity. But beware, you can do nasty things with git rebase, therefore stick to the simple procedure explained above to not corrupt your repository."},{"ref":"readme.html#license","title":"PropCheck - Property based testing for Elixir - License","type":"extras","doc":"PropCheck is provided under the GPL 3 License due to its intimate use of PropEr (which is licensed under GPL 3). In particular, PropEr&#39;s exclusion rule of open source software from copyleft applies here as well as described in this discussion on GitHub. Personally, I would prefer to use the LPGL to differentiate between extending PropEr and PropCheck as GPL-licensed software and the use of PropEr/PropCheck, which would not be infected by GPL&#39;s copyleft. But as long as PropEr does not change its licensing, we have no options. PropCheck is clearly an extension of PropEr, so it is derived work falling under GPL&#39;s copyleft. Using LGPL or any other license for PropCheck will not help, since GPL&#39;s copyleft overrules other licenses or result in an invalid or at least questionable licensing which does not help anybody. From my understanding of open source licenses as a legal amateur, the situation is currently as follows: Since PropCheck is a testing framework, the system under test is not infected by the CopyLeft of GPL, since PropCheck is only a tool used temporarily during development of the system under test. At least, if you don&#39;t distribute your system together with the test code and the test libs as a binary. Another friendly approach is to have the tests in a separate project, such that the tests are a real client of the system under test. But again, this is my personal take. In question, please consult a professional legal advisor."},{"ref":"changelog.html","title":"PropCheck Changelog","type":"extras","doc":"PropCheck Changelog"},{"ref":"changelog.html#1-2-3-development","title":"PropCheck Changelog - 1.2.3-Development","type":"extras","doc":"Support for parallel testing of state machines started. Requires also PRs in the upstream PropEr. An upgrade to (the future) PropEr 1.4 will result in better reports."},{"ref":"changelog.html#1-2-2-development","title":"PropCheck Changelog - 1.2.2-Development","type":"extras","doc":"Fixes to statemachine reporter by @aherranz and @devonestes Fixes to the Readme linking by @oo6"},{"ref":"changelog.html#1-2-1","title":"PropCheck Changelog - 1.2.1","type":"extras","doc":"Support for Elixir 1.10 in tests Refactorings of the statemachine implementation to be closer to PropEr. Thanks to https://github.com/x4lldux. property/1for marking properties to be implemented in the future. Thanks to https://github.com/evnu Prevent crashing if no counter examples was returned in a failing property. Thanks to https://github.com/evnu Consistent reporting of Erlang terms in Elixir syntax. Thanks to https://github.com/x4lldux. Enhanced handling and reporting of exception. Thanks to https://github.com/evnu Include credo in the build. Thanks to https://github.com/evnu Fix PROPCHECK_VERBOSE to work with property Allow PROPCHECK_VERBOSE=0 to make all properties quiet GitHub Actions are the new CI environemnt. Thanks to https://github.com/evnu Pinning of variables in let allows easier re-use of variables. Thanks to https://github.com/Ecialo"},{"ref":"changelog.html#1-2-0","title":"PropCheck Changelog - 1.2.0","type":"extras","doc":"Handling of tags corrected. This changes slighty existing the behavior and gives reason to introduce a new minor version. Thanks to https://github.com/evnu Verbose settings can be configured at the command line via environment variable PROPCHECK_VERBOSE. Thanks to https://github.com/evnu Setting default options to forall on module or describe level. Thanks to https://github.com/x4lldux. Support for Elixir 1.9 in tests. Thanks to https://github.com/evnu Moving back from CircleCI to TravisCI. Thanks to https://github.com/evnu"},{"ref":"changelog.html#1-1-5","title":"PropCheck Changelog - 1.1.5","type":"extras","doc":":verbose option is propagated from the property directly to forall. Thanks to https://github.com/evnu Storing of counter-examples can excluded by tag :store_counter_example Thanks to https://github.com/evnu Improved documentation for longer statemachine runs. Thanks to https://github.com/adkron Improved error message for missing command in StateM.DSL. Thanks to https://github.com/devonestes Introduction of linter credo and CircleCi as new CI tool. Thanks to https://github.com/evnu let syntax is the same now as forall. Thanks to https://github.com/evnu"},{"ref":"changelog.html#1-1-4","title":"PropCheck Changelog - 1.1.4","type":"extras","doc":"Fixes an issue with the setup of regular and targeted properties rendering 1.1.3 unusable Enhanced documentation for targeted properties"},{"ref":"changelog.html#1-1-3","title":"PropCheck Changelog - 1.1.3","type":"extras","doc":"Better command generator with improved shrinking for complex argument generations. Support for map-generator, thanks to https://github.com/IRog Support for targeted properties, a new feature of Proper 1.3 Requires at least Elixir 1.5"},{"ref":"changelog.html#1-1-2","title":"PropCheck Changelog - 1.1.2","type":"extras","doc":"Proper v1.3.0 is supported (effectively, all 1.x versions are allowed as dependency)"},{"ref":"changelog.html#1-1-1","title":"PropCheck Changelog - 1.1.1","type":"extras","doc":"the weight callback for the DSL was incorrectly specified and documented. Thanks to https://github.com/adkron"},{"ref":"changelog.html#1-1-0","title":"PropCheck Changelog - 1.1.0","type":"extras","doc":"New command oriented DSL for testing stateful systems, inspired by EQC and discussions about stateful testing in StreamData More details regarding licensing Rerun of properties fixed Better and corrected type specs, compatible with dialyxir 1.0.0(-rc*) Old modules for automatic type generators removed. They were never completed and since 2016 no longer part of the API (i.e. even before release 0.0.1)."},{"ref":"changelog.html#1-0-6","title":"PropCheck Changelog - 1.0.6","type":"extras","doc":"After a counter example is resolved, the entire property is run again to ensure that no other counter examples exist. Thanks to https://github.com/evnu tabs vs whitespace corrected for test cases, thanks to https://github.com/ryanwinchester added a hint about stored counterexamples for users, thanks to https://github.com/evnu Corrected formatting of markdown for documentation, thanks to https://github.com/zamith"},{"ref":"changelog.html#1-0-5","title":"PropCheck Changelog - 1.0.5","type":"extras","doc":"Allows to use ExUnit assertions as boolean conditions, thanks to https://github.com/evnu let and let_shrink allow more than two parameters, thanks to https://github.com/BinaryNoggin Errors, that aren&#39;t counter examples, are no longer stored as counter examples, thanks to https://github.com/evnu new feature sample_shrink, thanks to https://github.com/evnu the examples for stateful testing use GenServer.stop/0 for a reliable stopping of gen servers. several documentation issues"},{"ref":"changelog.html#1-0-4","title":"PropCheck Changelog - 1.0.4","type":"extras","doc":"produce has now a valid default parameter Removed several lazy compiler warnings Link in README corrected."},{"ref":"changelog.html#1-0-3","title":"PropCheck Changelog - 1.0.3","type":"extras","doc":"Removed debug log output."},{"ref":"changelog.html#1-0-2","title":"PropCheck Changelog - 1.0.2","type":"extras","doc":"only labeled, never released..."},{"ref":"changelog.html#1-0-1","title":"PropCheck Changelog - 1.0.1","type":"extras","doc":"Bugfix for Mix integration in Umbrella projects, thanks to https://github.com/evnu"},{"ref":"changelog.html#1-0-0","title":"PropCheck Changelog - 1.0.0","type":"extras","doc":"Counter examples are automatically stored and reapplied until the properties work or the counter examples are deleted. See https://github.com/alfert/propcheck/pull/18 Mix configuration for counter examples file and for inspecting and cleaning counter examples."},{"ref":"changelog.html#0-0-2","title":"PropCheck Changelog - 0.0.2","type":"extras","doc":"Fixed a lot of 1.5 (and 1.4) Elixir warnings thanks to https://github.com/evnu Readme additions regarding installation thanks to https://github.com/evnu Added more concurrency robustness for the ping pong tests Fixed a bug a in the movie server, which did not startup properly."},{"ref":"changelog.html#0-0-1","title":"PropCheck Changelog - 0.0.1","type":"extras","doc":"Initial release"}]